import copy
import unittest
import warnings

import numpy as np
import pytest

import skrf as rf
from skrf import two_port_reflect
from skrf.calibration import (
    PHN,
    SOLT,
    MultiportCal,
    MultiportSOLT,
    NISTMultilineTRL,
    TUGMultilineTRL,
    TwelveTerm,
    TwoPortOnePath,
    UnknownThru,
    compute_switch_terms,
    determine_line,
    determine_reflect,
    terminate,
    terminate_nport,
)
from skrf.media import Coaxial, DistributedCircuit
from skrf.networkSet import NetworkSet
from skrf.util import suppress_warning_decorator

# number of frequency points to test calibration at .
# i choose 1 for speed, but given that many tests employ *random*
# networks values >100 are better for initial verification
global NPTS
NPTS = 1

# WR10/WG27/R900 75 to 110 GHz, 0.1x0.05 inch (2.54x1.27 mm)
# z0 from 610 to 446 ohm
WG_lossless = rf.RectangularWaveguide(rf.F(75, 100, NPTS, unit='GHz'),
                                      a=100*rf.mil, z0_override=50)
WG = rf.RectangularWaveguide(rf.F(75, 100, NPTS, unit='GHz'), a=100*rf.mil,
                                      rho='gold', z0_override=50)

def _compare_dicts_allclose(first: dict, second: dict) -> None:
    assert first.keys() == second.keys()
    for k in first.keys():
        np.testing.assert_allclose(first[k], second[k], err_msg=f"Values from key '{k}' not equal!")


class DetermineTest(unittest.TestCase):
    def setUp(self):
        self.wg = WG
        wg = self.wg
        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')


        #thru
        self.T = wg.thru()
        self.T_m = self.embed(self.T)

        #line
        self.L = wg.line(80,'deg')
        self.L_approx = self.wg.line(90,'deg')
        self.L_m = self.embed(self.L)

        # reflect
        self.r = [ wg.delay_load(p,k,'deg') \
                    for k in [-10,10,88,92] \
                    for p in [-.9,-1]]


        self.R = [rf.two_port_reflect(k,k) for k in self.r]

        short= wg.short()
        open = wg.open()
        self.r_estimate = [short, short, short, short, open, open, open ,open]
        self.R_m = [self.embed(k) for k in self.R]

    def embed(self,x):
        return self.X**x**self.Y

    def test_determine_line(self):
        L_found = determine_line(self.T_m, self.L_m,
                                 line_approx=self.L_approx)
        self.assertEqual(L_found,self.L)

    def test_determine_reflect(self):
        r_found = [determine_reflect(self.T_m,k,self.L_m, l, self.L_approx) \
                   for k,l in zip(self.R_m, self.r_estimate)]

        [ self.assertEqual(k,l) for k,l in zip(self.r, r_found)]

    def test_determine_reflect_matched_thru_and_line_ideal_reflect(self):
        freq = rf.F(.25, .7, 21, unit = 'GHz')
        medium = Coaxial.from_attenuation_VF(freq, att = 3.0, VF = .69)

        thru = medium.line(0, 'm')
        line = medium.line(0.12, 'm')

        short = rf.two_port_reflect(medium.short(), medium.short())
        r = determine_reflect(thru, short, line)
        np.testing.assert_array_almost_equal( r.s, -np.ones_like(r.s))

        reflect = rf.two_port_reflect(medium.open(), medium.open())
        r = determine_reflect(thru, reflect, line, reflect_approx = medium.open())
        np.testing.assert_array_almost_equal(r.s, np.ones_like(r.s))

    def test_determine_reflect_matched_thru_and_line(self):
        freq = rf.F(.25, .7, 40, unit = 'GHz')
        medium = Coaxial.from_attenuation_VF(freq, att = 3.0, VF = .69)

        thru = medium.line(0, 'm')
        line = medium.line(0.12, 'm')
        short = medium.short()

        rng = np.random.default_rng(12)
        short.s[:, 0, 0] += rng.uniform(-.02, 0.02, freq.f.size) + rng.uniform(-.02, 0.02, freq.f.size)*1j

        r = determine_reflect(thru, rf.two_port_reflect(short, short), line)
        np.testing.assert_array_almost_equal( r.s, short.s)


    def test_determine_reflect_regression(self):
        # this test case fails with only regularization on the t-parameters of thru ** line.inv, see gh-870
        freq = rf.F(434615384.6153846e-9, .7, 1, unit = 'GHz')
        medium = Coaxial.from_attenuation_VF(freq, att = 3.0, VF = .69)
        thru= medium.line(0, 'm')
        line = medium.line(0.12, 'm')
        short = rf.Network(frequency=freq, s=[[[(-1.+0.017870117714376983j)] ] ])

        r = determine_reflect(thru, rf.two_port_reflect(short, short), line)
        np.testing.assert_array_almost_equal( r.s, short.s)

class ComputeSwitchTermsTest(unittest.TestCase):
    '''
    test the indirect method of computing the switch terms
    with at least three reciprocal devices
    '''
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg = self.wg

        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')

        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        # Reciprocal devices: asymmetric and both transmissive and reflective devices
        Rs = [25, 50, 100, ]
        stands = [wg.resistor(R)**wg.shunt_resistor(R) for R in Rs]
        stands_meas = [terminate(self.X**k**self.Y, self.gamma_f, self.gamma_r) for k in stands]

        self.gamma_f_indirect, self.gamma_r_indirect = compute_switch_terms(stands_meas)

    def test_gamma_f(self):
        self.assertTrue(all(np.abs(self.gamma_f_indirect.s - self.gamma_f.s) < 1e-9))

    def test_gamma_r(self):
        self.assertTrue(all(np.abs(self.gamma_r_indirect.s - self.gamma_r.s) < 1e-9))

class CalibrationTest:
    """
    This is the generic Calibration test case which all Calibration
    Subclasses should be able to pass. They must implement
    """

    def test_accuracy_of_dut_correction(self):
        a = self.wg.random(n_ports=self.n_ports, name = 'actual')
        m = self.measure(a)
        c = self.cal.apply_cal(m)
        c.name = 'corrected'
        self.assertEqual(c,a)

    def test_error_ntwk(self):
        a = self.cal.error_ntwk

    def test_coefs_ntwks(self):
        a = self.cal.coefs_ntwks

    @suppress_warning_decorator("only gave a single measurement orientation")
    def test_caled_ntwks(self):
        a = self.cal.caled_ntwks

    @suppress_warning_decorator("only gave a single measurement orientation")
    def test_residual_ntwks(self):
        a = self.cal.residual_ntwks

    def test_embed_then_apply_cal(self):
        a = self.wg.random(n_ports=self.n_ports)
        self.assertEqual(self.cal.apply_cal(self.cal.embed(a)),a)

    def test_embed_equal_measure(self):
        a = self.wg.random(n_ports=self.n_ports)
        self.assertEqual(self.cal.embed(a),self.measure(a))

    @suppress_warning_decorator("n_thrus is None")
    def test_from_coefs(self):
        cal_from_coefs = self.cal.from_coefs(self.cal.frequency, self.cal.coefs)
        ntwk = self.wg.random(n_ports=self.n_ports)
        self.assertEqual(cal_from_coefs.apply_cal(self.cal.embed(ntwk)),ntwk)

    @suppress_warning_decorator("n_thrus is None")
    def test_from_coefs_ntwks(self):
        cal_from_coefs = self.cal.from_coefs_ntwks(self.cal.coefs_ntwks)
        ntwk = self.wg.random(n_ports=self.n_ports)
        self.assertEqual(cal_from_coefs.apply_cal(self.cal.embed(ntwk)),ntwk)


class CalibrationInputsTest(unittest.TestCase):
    """Test Calibration class input validation"""
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg = self.wg

        self.ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.thru(name='thru'),
            ]

        self.measured = [k.copy() for k in self.ideals]

    def init(self):
        self.cal = rf.Calibration(
            ideals = self.ideals,
            measured = self.measured
            )

    def test_meas_z0(self):
        self.measured[0].z0 *= 0.5
        with pytest.raises(ValueError):
            self.init()

    def test_ideals_z0(self):
        self.ideals[0].z0 *= 0.5
        with pytest.raises(ValueError):
            self.init()

    def test_ideals_frequency(self):
        freqs = rf.F(1, 10, NPTS+1, unit="GHz")
        self.ideals[-1] = rf.Network(s=np.zeros([NPTS + 1, 2, 2]), frequency=freqs)
        with pytest.raises(IndexError):
            self.init()

    def test_measured_frequency(self):
        freqs = rf.F(1, 10, NPTS+1, unit="GHz")
        self.measured[-1] = rf.Network(s=np.zeros([NPTS + 1, 2, 2]), frequency=freqs)
        with pytest.raises(ValueError):
            self.init()

    def test_ideals_frequency2(self):
        freqs = self.wg.frequency.copy()
        freqs._f += 1e9
        self.ideals[-1].frequency = freqs
        with pytest.raises(IndexError):
            self.init()

    def test_measured_frequency2(self):
        freqs = self.wg.frequency.copy()
        freqs._f += 1e9
        self.measured[-1].frequency = freqs
        with pytest.raises(ValueError):
            self.init()

class OnePortTest(unittest.TestCase, CalibrationTest):
    """
    One-port calibration test.


    """
    def setUp(self):
        self.n_ports = 1
        self.wg = WG
        wg = self.wg

        self.E = wg.random(n_ports =2, name = 'E')

        ideals = [
                wg.short( name='short'),
                wg.delay_short( 45.,'deg',name='ew'),
                wg.delay_short( 90.,'deg',name='qw'),
                wg.match( name='load'),
                ]
        measured = [self.measure(k) for k in ideals]

        self.cal = rf.OnePort(
            is_reciprocal = True,
            ideals = ideals,
            measured = measured,
            )

    def test_input_networks_1port(self):
        # test users do not enter 2-port networks by accident
        with self.assertRaises(RuntimeError):
            wg = self.wg
            ideals = [
                    two_port_reflect(wg.short( name='short'), wg.short( name='short')),
                    wg.delay_short( 45.,'deg',name='ew'),
                    wg.delay_short( 90.,'deg',name='qw'),
                    wg.match( name='load'),
                    ]
            measured = [self.measure(k) for k in ideals]
            cal = rf.OnePort(
               is_reciprocal = True,
               ideals = ideals,
               measured = measured,
               )
            cal.run()

    def measure(self, ntwk):
        out = self.E**ntwk
        out.name = ntwk.name
        return out

    def test_accuracy_of_directivity(self):
        self.assertEqual(
            self.E.s11,
            self.cal.coefs_ntwks['directivity'],
            )

    def test_accuracy_of_source_match(self):
        self.assertEqual(
            self.E.s22,
            self.cal.coefs_ntwks['source match'],
            )

    def test_accuracy_of_reflection_tracking(self):
        self.assertEqual(
            self.E.s21*self.E.s12,
            self.cal.coefs_ntwks['reflection tracking'],
            )


class SDDLTest(OnePortTest):
    def setUp(self):
        self.n_ports = 1
        #Exact only with a lossless waveguide
        self.wg = WG_lossless
        wg = self.wg

        self.E = wg.random(n_ports =2, name = 'E')
        #self.E.s[0,:,:] = np.array([[.1j,1],[1j,1j+2]])
        #print self.E.s[0]

        ideals = [
                wg.short( name='short'),
                wg.delay_short( 45.,'deg',name='ew'),
                wg.delay_short( 90.,'deg',name='qw'),
                wg.load(.2+.2j, name='load'),
                ]
        actuals = [
                wg.short( name='short'),
                wg.delay_short( 10.,'deg',name='ew'),
                wg.delay_short( 33.,'deg',name='qw'),
                wg.load(.2+.2j, name='load'),
                ]
        measured = [self.measure(k) for k in actuals]

        self.cal = rf.SDDL(
            is_reciprocal = True,
            ideals = ideals,
            measured = measured,
            )

    def test_init_with_nones(self):
        wg=self.wg
        wg.frequency = rf.F.from_f([100], unit='GHz')

        self.E = wg.random(n_ports =2, name = 'E')

        ideals = [
                wg.short( name='short'),
                None,
                None,
                wg.load(.2+.2j, name='load'),
                ]
        actuals = [
                wg.short( name='short'),
                wg.delay_short( 10.,'deg',name='ew'),
                wg.delay_short( 33.,'deg',name='qw'),
                wg.load(.2+.2j, name='load'),
                ]
        measured = [self.measure(k) for k in actuals]

        self.cal = rf.SDDL(
            is_reciprocal = True,
            ideals = ideals,
            measured = measured,
            )
        self.cal.run()

    @pytest.mark.skip(reason='not applicable ')
    def test_from_coefs(self):
        pass

    @pytest.mark.skip(reason='not applicable ')
    def test_from_coefs_ntwks(self):
        pass


class SDDLWeikle(OnePortTest):
    def setUp(self):
        self.n_ports = 1
        #Exact only with a lossless waveguide
        self.wg = WG_lossless
        wg = self.wg
        self.E = wg.random(n_ports =2, name = 'E')
        #self.E.s[0,:,:] = np.array([[.1j,1],[1j,1j+2]])
        #print self.E.s[0]

        ideals = [
                wg.short( name='short'),
                wg.delay_short( 45.,'deg',name='ew'),
                wg.delay_short( 90.,'deg',name='qw'),
                wg.load(.2+.2j, name='load'),
                ]
        actuals = [
                wg.short( name='short'),
                wg.delay_short( 10.,'deg',name='ew'),
                wg.delay_short( 80.,'deg',name='qw'),
                wg.load(.2+.2j, name='load'),
                ]
        measured = [self.measure(k) for k in actuals]

        self.cal = rf.SDDLWeikle(
            is_reciprocal = True,
            ideals = ideals,
            measured = measured,
            )

    @pytest.mark.skip(reason='not applicable ')
    def test_from_coefs(self):
        pass

    @pytest.mark.skip(reason='not applicable ')
    def test_from_coefs_ntwks(self):
        pass


class SDDMTest(OnePortTest):
    """
    This is a specific test of SDDL to verify it works when the load is
    a matched load. This test has been used to show that the SDDLWeikle
    variant fails, with a perfect matched load.
    """
    def setUp(self):

        self.n_ports = 1
        #Exact only with a lossless waveguide
        self.wg = WG_lossless
        wg = self.wg

        self.E = wg.random(n_ports =2, name = 'E')

        ideals = [
                wg.short( name='short'),
                wg.delay_short( 45.,'deg',name='ew'),
                wg.delay_short( 90.,'deg',name='qw'),
                wg.match( name='load'),
                ]
        actuals = [
                wg.short( name='short'),
                wg.delay_short( 10.,'deg',name='ew'),
                wg.delay_short( 80.,'deg',name='qw'),
                wg.match(name='load'),
                ]
        measured = [self.measure(k) for k in actuals]

        self.cal = rf.SDDL(
            is_reciprocal = True,
            ideals = ideals,
            measured = measured,
            )

    @pytest.mark.skip(reason='not applicable ')
    def test_from_coefs(self):
        pass

    @pytest.mark.skip(reason='not applicable ')
    def test_from_coefs_ntwks(self):
        pass


@pytest.mark.skip()
class PHNTest(OnePortTest):
    """
    """
    def setUp(self):

        self.n_ports = 1
        self.wg = WG
        wg = self.wg

        self.E = wg.random(n_ports =2, name = 'E')
        known1 = wg.random()
        known2 = wg.random()
        #known1 = wg.short()
        #known2 = wg.load(rand() + rand()*1j)

        ideals = [
                wg.delay_short( 45.,'deg',name='ideal ew'),
                wg.delay_short( 90.,'deg',name='ideal qw'),
                known1,
                known2,
                ]
        actuals = [
                wg.delay_short( 33.,'deg',name='true ew'),
                wg.delay_short( 110.,'deg',name='true qw'),
                known1,
                known2,
                ]
        measured = [self.measure(k) for k in actuals]
        self.actuals = actuals

        self.cal = PHN(
            is_reciprocal = True,
            ideals = ideals,
            measured = measured,
            )


    def test_determine_ideals(self):
        self.cal.run()

        self.assertEqual(self.actuals[0], self.cal.ideals[0])
        self.assertEqual(self.actuals[1], self.cal.ideals[1])

    @pytest.mark.skip(reason='not applicable ')
    def test_from_coefs(self):
        pass

    @pytest.mark.skip(reason='not applicable ')
    def test_from_coefs_ntwks(self):
        pass


class EightTermTest(unittest.TestCase, CalibrationTest):
    def setUp(self):
        self.n_ports = 2
        self.wg =WG
        wg= self.wg

        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        #Isolation terms
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.thru(name='thru'),
            ]

        measured = [self.measure(k) for k in ideals]

        self.cal = rf.EightTerm(
            ideals = ideals,
            measured = measured,
            isolation = measured[2],
            switch_terms = (self.gamma_f, self.gamma_r)
            )

    def terminate(self, ntwk):
        """
        terminate a measured network with the switch terms
        """
        return terminate(ntwk,self.gamma_f, self.gamma_r)

    def measure(self,ntwk):
        out =  self.terminate(self.X**ntwk**self.Y)
        out.name = ntwk.name
        out.s[:,1,0] += self.If.s[:,0,0]
        out.s[:,0,1] += self.Ir.s[:,0,0]
        return out

    def test_unterminating(self):
        a = self.wg.random(n_ports=self.n_ports)
        #unterminated measurement
        ut =  self.X**a**self.Y
        #terminated measurement
        m = self.measure(a)
        #Remove leakage
        m.s[:,1,0] -= self.If.s[:,0,0]
        m.s[:,0,1] -= self.Ir.s[:,0,0]
        self.assertEqual(self.cal.unterminate(m), ut)

    def test_forward_directivity_accuracy(self):
        self.assertEqual(
            self.X.s11,
            self.cal.coefs_ntwks['forward directivity'])

    def test_forward_source_match_accuracy(self):
        self.assertEqual(
            self.X.s22 ,
            self.cal.coefs_ntwks['forward source match'] )

    def test_forward_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.X.s21 * self.X.s12 ,
            self.cal.coefs_ntwks['forward reflection tracking'])

    def test_reverse_source_match_accuracy(self):
        self.assertEqual(
            self.Y.s11 ,
            self.cal.coefs_ntwks['reverse source match']   )

    def test_reverse_directivity_accuracy(self):
        self.assertEqual(
            self.Y.s22 ,
            self.cal.coefs_ntwks['reverse directivity']  )

    def test_reverse_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.Y.s21 * self.Y.s12 ,
            self.cal.coefs_ntwks['reverse reflection tracking'])

    def test_k_accuracy(self):
        self.assertEqual(
            self.X.s21/self.Y.s12 ,
            self.cal.coefs_ntwks['k']  )

    def test_forward_isolation_accuracy(self):
        self.assertEqual(
            self.If.s11 ,
            self.cal.coefs_ntwks['forward isolation']  )

    def test_reverse_isolation_accuracy(self):
        self.assertEqual(
            self.Ir.s11 ,
            self.cal.coefs_ntwks['reverse isolation']  )

    def test_verify_12term(self):
        self.assertTrue(self.cal.verify_12term_ntwk.s_mag.max() < 1e-3)

    def test_coefs_8term(self):
        self.assertEqual(self.cal.coefs_8term, self.cal.coefs)

    def test_coefs_12term(self):
        _compare_dicts_allclose(self.cal.coefs_12term, rf.convert_8term_2_12term(self.cal.coefs))

    def test_input_not_modified(self):
        dut_before = self.X.copy()
        self.cal.apply_cal(self.X)
        self.assertEqual(dut_before, self.X)

    def test_convert_8term_2_12term_subtests(self):

        for st in [
            (self.cal.coefs['forward switch term'], self.cal.coefs['reverse switch term']),
            (np.zeros(self.cal.frequency.npoints, dtype=complex), np.zeros(self.cal.frequency.npoints, dtype=complex)),
            (self.cal.coefs['forward switch term'], np.zeros(self.cal.frequency.npoints, dtype=complex)),
            (np.zeros(self.cal.frequency.npoints, dtype=complex), self.cal.coefs['reverse switch term']),
        ]:
            with self.subTest(st=st):
                coefs = copy.deepcopy(self.cal.coefs)

                coefs['forward switch term'] = st[0]
                coefs['reverse switch term'] = st[1]

                coefs_12term = rf.convert_8term_2_12term(coefs)
                coefs_8term = rf.convert_12term_2_8term(coefs_12term)
                for k in self.cal.coefs.keys():
                    np.testing.assert_almost_equal(coefs_8term[k], coefs[k])

class TRLTest(EightTermTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg= self.wg


        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')
        # make error networks have s21,s12 >> s11,s22 so that TRL
        # can guess at line length
        self.X.s[:,0,0] *=1e-1
        self.Y.s[:,0,0] *=1e-1
        self.X.s[:,1,1] *=1e-1
        self.Y.s[:,1,1] *=1e-1

        actuals = [
            wg.thru( name='thru'),
            rf.two_port_reflect(wg.load(-.9-.1j),wg.load(-.9-.1j)),
            wg.attenuator(-3,True, 45,'deg')
            #wg.line(45,'deg',name='line'),
            ]
        self.actuals=actuals
        ideals = [
            wg.thru( name='thru'),
            wg.short(nports=2, name='short'),
            wg.line(90,'deg',name='line'),
            ]

        measured = [self.measure(k) for k in actuals]

        self.cal = rf.TRL(
            ideals = ideals,
            measured = measured,
            isolation = measured[1],
            switch_terms = (self.gamma_f, self.gamma_r)
            )

    def test_found_line(self):
        self.cal.run()
        self.assertTrue(self.cal.ideals[2]==self.actuals[2])

    def test_found_reflect(self):
        self.cal.run()
        self.assertTrue(self.cal.ideals[1]==self.actuals[1])

class TRLLongThruTest(EightTermTest):
    """
    Test TRL calibration with non-zero length thru.
    In this case calibration is done at thru center.
    """
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg = self.wg

        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')
        # make error networks have s21,s12 >> s11,s22 so that TRL
        # can guess at line length
        self.X.s[:,0,0] *= 1e-1
        self.Y.s[:,0,0] *= 1e-1
        self.X.s[:,1,1] *= 1e-1
        self.Y.s[:,1,1] *= 1e-1

        # Reflect reference plane is at the thru center
        self.reflect = wg.load(-.9-.1j)
        reflect_shifted = wg.line(50, 'um') ** self.reflect

        actuals = [
            wg.line(100, 'um', name='thru'),
            rf.two_port_reflect(reflect_shifted, reflect_shifted),
            wg.line(1100, 'um', name='thru'),
            ]

        self.actuals = actuals

        ideals = [
            actuals[0],
            wg.short(nports=2, name='short'),
            actuals[2]
            ]

        measured = [self.measure(k) for k in actuals]

        # Calibration is done at the center of the thru
        # Add half thru to error networks so that tests pass
        self.X = self.X ** wg.line(50, 'um')
        self.Y = wg.line(50, 'um') ** self.Y

        self.cal = rf.TRL(
            ideals = ideals,
            measured = measured,
            isolation = measured[1],
            switch_terms = (self.gamma_f, self.gamma_r)
            )

    def test_found_line(self):
        self.cal.run()
        # Solved line is difference between line and thru
        self.assertTrue(self.cal.ideals[2]==self.actuals[2] ** self.actuals[0].inv)

    def test_found_reflect(self):
        self.cal.run()
        # Solved reflect is at the thru center
        self.assertTrue(self.cal.ideals[1]==rf.two_port_reflect(self.reflect, self.reflect))

class TRLWithNoIdealsTest(EightTermTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg= self.wg

        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')
        # make error networks have s21,s12 >> s11,s22 so that TRL
        # can guess at line length
        self.X.s[:,0,0] *=1e-1
        self.Y.s[:,0,0] *=1e-1
        self.X.s[:,1,1] *=1e-1
        self.Y.s[:,1,1] *=1e-1

        r = wg.delay_short(20,'deg')

        self.actuals=[wg.thru( name='thru'),
                      rf.two_port_reflect(r,r),\
                      wg.attenuator(-3,True, 45,'deg')]

        measured = [self.measure(k) for k in self.actuals]

        self.cal = rf.TRL(
            ideals = None,
            measured = measured,
            isolation = measured[1],
            switch_terms = (self.gamma_f, self.gamma_r)
            )

    def test_found_line(self):
        self.cal.run()
        self.assertTrue(self.cal.ideals[2]==self.actuals[2])

    def test_found_reflect(self):
        self.cal.run()
        self.assertTrue(self.cal.ideals[1]==self.actuals[1])


class TRLMultiline(EightTermTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg= self.wg

        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')
        # make error networks have s21,s12 >> s11,s22 so that TRL
        # can guess at line length
        #self.X.s[:,0,0] *=1e-1
        #self.Y.s[:,0,0] *=1e-1
        #self.X.s[:,1,1] *=1e-1
        #self.Y.s[:,1,1] *=1e-1

        actuals = [
            wg.thru( name='thru'),
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.attenuator(-3,True, 45,'deg'),
            wg.attenuator(-6,True, 90,'deg'),
            wg.attenuator(-8,True, 145,'deg'),
            ]
        self.actuals=actuals

        measured = [self.measure(k) for k in actuals]

        self.cal = rf.TRL(
            ideals = [None, -1,1,None,None,None],
            measured = measured,
            isolation = measured[1],
            switch_terms = (self.gamma_f, self.gamma_r),
            n_reflects=2,
            )

    def test_found_line(self):
        self.cal.run()
        for k in range(2,5):
            self.assertTrue(self.cal.ideals[k]==self.actuals[k])


class NISTMultilineTRLTest(EightTermTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg= self.wg

        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name = 'Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')

        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        actuals = [
            wg.thru(),
            rf.two_port_reflect(wg.load(-.98-.1j),wg.load(-.98-.1j)),
            rf.two_port_reflect(wg.load(.99+0.05j),wg.load(.99+0.05j)),
            wg.line(100,'um'),
            wg.line(200,'um'),
            wg.line(900,'um'),
            ]

        self.actuals=actuals

        measured = [self.measure(k) for k in actuals]

        self.cal = NISTMultilineTRL(
            measured = measured,
            isolation = measured[1],
            Grefls = [-1, 1],
            l = [0, 100e-6, 200e-6, 900e-6],
            er_est = 1,
            switch_terms = (self.gamma_f, self.gamma_r),
            gamma_root_choice = 'real'
            )

    def test_gamma(self):
        self.assertTrue(max(np.abs(self.wg.gamma-self.cal.gamma)) < 1e-3)


class NISTMultilineTRLTest2(NISTMultilineTRLTest):
    """ Test characteristic impedance change and reference plane shift.
    """
    def setUp(self):
        global NPTS
        self.n_ports = 2
        self.wg = WG
        wg = self.wg

        rng = np.random.default_rng()

        r = rng.uniform(10,100,NPTS)
        l = 1e-9*rng.uniform(100,200,NPTS)
        g = np.zeros(NPTS)
        c = 1e-12*rng.uniform(100,200,NPTS)

        rlgc = DistributedCircuit(frequency=wg.frequency, z0=None, R=r, L=l, G=g, C=c)
        self.rlgc = rlgc

        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name = 'Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        actuals = [
            rlgc.line(1000,'um'),
            rlgc.line(500,'um') ** rlgc.short(nports=2) ** rlgc.line(500,'um'),
            rlgc.line(1010,'um'),
            rlgc.line(1100,'um'),
            rlgc.line(1800,'um'),
            ]

        self.actuals=actuals

        measured = [self.measure(k) for k in actuals]
        self.measured = measured

        self.cal = NISTMultilineTRL(
            measured = measured,
            isolation = measured[1],
            Grefls = [-1],
            refl_offset = 500e-6,
            l = [1000e-6, 1010e-6, 1100e-6, 1800e-6],
            switch_terms = (self.gamma_f, self.gamma_r),
            c0=c,
            z0_ref=50,
            gamma_root_choice = 'real'
            )

        self.cal_shift = NISTMultilineTRL(
            measured = measured,
            isolation = measured[1],
            Grefls = [-1],
            refl_offset = 0,
            l = [0, 10e-6, 100e-6, 800e-6],
            switch_terms = (self.gamma_f, self.gamma_r),
            ref_plane = -500e-6,
            c0=c,
            z0_ref=50,
            gamma_root_choice = 'real'
            )

    def test_gamma(self):
        self.assertTrue(max(np.abs(self.rlgc.gamma-self.cal.gamma)) < 1e-3)

    def test_z0(self):
        self.assertTrue(max(np.abs(self.rlgc.z0-self.cal.z0)) < 1e-3)

    def test_shift(self):
        for k in self.cal.coefs.keys():
            self.assertTrue(all(np.abs(self.cal.coefs[k] - self.cal_shift.coefs[k]) < 1e-9))


    def test_numpy_float_arguments(self):
        # see gh-895
        cal = NISTMultilineTRL(
            measured = self.measured[:3],
            Grefls = [-1],
            l = [np.float64(1000e-6), 1010e-6],
            switch_terms = (self.gamma_f, self.gamma_r),
            )
        cal.run()
        cal.apply_cal(self.measured[0])

        cal = NISTMultilineTRL(
            measured = self.measured[:3],
            Grefls = [-1],
            l = [1000e-6, 1010e-6],
            z0_ref = np.float64(50),
            z0_line = np.float64(50),
            switch_terms = (self.gamma_f, self.gamma_r),
            )
        cal.run()
        cal.apply_cal(self.measured[0])

class TUGMultilineTest(EightTermTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg= self.wg

        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name = 'Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')

        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        actuals = [
            wg.thru(),
            rf.two_port_reflect(wg.load(-.98-.1j),wg.load(-.98-.1j)),
            rf.two_port_reflect(wg.load(.99+0.05j),wg.load(.99+0.05j)),
            wg.line(100,'um'),
            wg.line(200,'um'),
            wg.line(900,'um'),
            ]

        self.actuals=actuals

        measured = [self.measure(k) for k in actuals]

        self.cal = TUGMultilineTRL(
            line_meas = [measured[0]] + measured[3:],
            line_lengths = [0, 100e-6, 200e-6, 900e-6],
            er_est = 1,
            reflect_meas = measured[1:3],
            reflect_est = [-1, 1],
            isolation = measured[1],
            switch_terms = (self.gamma_f, self.gamma_r)
            )

    def test_gamma(self):
        self.assertTrue(max(np.abs(self.wg.gamma-self.cal.gamma)) < 1e-3)

@pytest.mark.skip()
class TREightTermTest(unittest.TestCase, CalibrationTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg= self.wg


        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')


        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.thru(name='thru'),
            ]

        measured = [self.measure_std(k) for k in ideals]

        cal1 = rf.TwoPortOnePath(
            ideals = ideals,
            measured = measured
            )
        switch_terms = (cal1.coefs_ntwks['forward switch term'],
                        cal1.coefs_ntwks['reverse switch term'])


        measured = [self.measure(k) for k in ideals]
        self.cal = rf.EightTerm(
            ideals = ideals,
            measured = measured,
            switch_terms = switch_terms,
            )
        raise ValueError()

    def measure_std(self,ntwk):
        r= self.wg.random(2)
        m = ntwk.copy()
        mf = self.X**ntwk**self.Y

        m.s[:,1,0] = mf.s[:,1,0]
        m.s[:,0,0] = mf.s[:,0,0]
        m.s[:,1,1] = r.s[:,1,1]
        m.s[:,0,1] = r.s[:,0,1]
        return m
    def measure(self,ntwk):

        m = ntwk.copy()
        mf = self.X**ntwk**self.Y
        mr = self.X**ntwk.flipped()**self.Y

        m.s[:,1,0] = mf.s[:,1,0]
        m.s[:,0,0] = mf.s[:,0,0]
        m.s[:,1,1] = mr.s[:,0,0]
        m.s[:,0,1] = mr.s[:,1,0]
        return m


class TwelveTermTest(unittest.TestCase, CalibrationTest):
    """
    This test verifies the accuracy of the SOLT calibration. Generating
    measured networks requires different error networks for forward and
    reverse excitation states, these are described as follows

    forward excitation
        used for S21 and S11
        Mf = Xf ** S ** Yf

    reverse excitation
        used for S12 and S22
        Mr = Xr ** S ** Yr


    """
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg  = self.wg
        self.Xf = wg.random(n_ports =2, name = 'Xf')
        self.Xr = wg.random(n_ports =2, name = 'Xr')
        self.Yf = wg.random(n_ports =2, name='Yf')
        self.Yr = wg.random(n_ports =2, name='Yr')
        self.If = wg.random(n_ports =1, name='If')
        self.Ir = wg.random(n_ports =1, name='Ir')

        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.random(2,name='rand1'),
            wg.random(2,name='rand2'),
            ]

        measured = [ self.measure(k) for k in ideals]

        self.cal = rf.TwelveTerm(
            ideals = ideals,
            measured = measured,
            n_thrus=2,
            isolation=measured[2]
            )

    def measure(self,ntwk):
        m = ntwk.copy()
        mf = self.Xf**ntwk**self.Yf
        mr = self.Xr**ntwk**self.Yr
        m.s[:,1,0] = mf.s[:,1,0]
        m.s[:,0,0] = mf.s[:,0,0]
        m.s[:,0,1] = mr.s[:,0,1]
        m.s[:,1,1] = mr.s[:,1,1]

        #Leakage
        m.s[:,1,0] += self.If.s[:,0,0]
        m.s[:,0,1] += self.Ir.s[:,0,0]
        return m

    def test_forward_directivity_accuracy(self):
        self.assertEqual(
            self.Xf.s11,
            self.cal.coefs_ntwks['forward directivity'])

    def test_forward_source_match_accuracy(self):
        self.assertEqual(
            self.Xf.s22 ,
            self.cal.coefs_ntwks['forward source match'] )

    def test_forward_load_match_accuracy(self):
        self.assertEqual(
            self.Yf.s11 ,
            self.cal.coefs_ntwks['forward load match'])

    def test_forward_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.Xf.s21 * self.Xf.s12 ,
            self.cal.coefs_ntwks['forward reflection tracking'])

    def test_forward_transmission_tracking_accuracy(self):
        self.assertEqual(
            self.Xf.s21*self.Yf.s21 ,
            self.cal.coefs_ntwks['forward transmission tracking'])

    def test_reverse_source_match_accuracy(self):
        self.assertEqual(
            self.Yr.s11 ,
            self.cal.coefs_ntwks['reverse source match']   )

    def test_reverse_directivity_accuracy(self):
        self.assertEqual(
            self.Yr.s22 ,
            self.cal.coefs_ntwks['reverse directivity']  )

    def test_reverse_load_match_accuracy(self):
        self.assertEqual(
            self.Xr.s22 ,
            self.cal.coefs_ntwks['reverse load match'])

    def test_reverse_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.Yr.s21 * self.Yr.s12 ,
            self.cal.coefs_ntwks['reverse reflection tracking'])

    def test_reverse_transmission_tracking_accuracy(self):
        self.assertEqual(
            self.Yr.s12*self.Xr.s12 ,
            self.cal.coefs_ntwks['reverse transmission tracking'])

    def test_forward_isolation_accuracy(self):
        self.assertEqual(
            self.If.s11,
            self.cal.coefs_ntwks['forward isolation'])

    def test_reverse_isolation_accuracy(self):
        self.assertEqual(
            self.Ir.s11,
            self.cal.coefs_ntwks['reverse isolation'])

    def test_convert_12term_2_8term(self):
        converted = rf.convert_8term_2_12term(
                    rf.convert_12term_2_8term(self.cal.coefs, redundant_k=True))

        for k in converted:
            self.assertTrue(abs(self.cal.coefs[k] - converted[k])<1e-9)

    def test_convert_12term_2_8term_correction_accuracy(self):
        converted = rf.convert_8term_2_12term(
                    rf.convert_12term_2_8term(self.cal.coefs, redundant_k=True))

        self.cal._coefs = converted
        a = self.wg.random(n_ports=2)
        m = self.measure(a)
        c = self.cal.apply_cal(m)

        self.assertEqual(a,c)

    def test_coefs_8term(self):
        self.assertEqual(self.cal.coefs_8term, rf.convert_12term_2_8term(self.cal.coefs))

    def test_coefs_12term(self):
        self.assertEqual(self.cal.coefs_12term, self.cal.coefs)


class TwelveTermSloppyInitTest(TwelveTermTest):
    """
    Test the TwelveTerm.__init__'s ability to
    1) determine the number of thrus (n_thrus) heuristically
    2) put the standards in correct order if they use sloppy_input

    It must be a entirely separate test because we want to ensure it
    creates an accurate calibration.
    """
    @suppress_warning_decorator("dictionary passed, sloppy_input")
    @suppress_warning_decorator("n_thrus is None")
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg  = self.wg
        self.Xf = wg.random(n_ports =2, name = 'Xf')
        self.Xr = wg.random(n_ports =2, name = 'Xr')
        self.Yf = wg.random(n_ports =2, name='Yf')
        self.Yr = wg.random(n_ports =2, name='Yr')
        #No leakage as it can interfere with thru detection, which is done
        #based on S21 and S12
        self.If = wg.match(n_ports =1, name='If')
        self.Ir = wg.match(n_ports =1, name='Ir')

        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.attenuator(-20,name='atten'),
            wg.line(45,'deg',name = 'line') ,
            ]


        measured = [ self.measure(k) for k in ideals]

        self.cal= TwelveTerm(
            ideals = NetworkSet(ideals).to_dict(),
            measured = NetworkSet(measured).to_dict(),
            n_thrus=None,
            )

    def measure(self,ntwk):
        m = ntwk.copy()
        mf = self.Xf**ntwk**self.Yf
        mr = self.Xr**ntwk**self.Yr
        m.s[:,1,0] = mf.s[:,1,0]
        m.s[:,0,0] = mf.s[:,0,0]
        m.s[:,0,1] = mr.s[:,0,1]
        m.s[:,1,1] = mr.s[:,1,1]
        return m


class SOLTTest(TwelveTermTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg  = self.wg
        self.Xf = wg.random(n_ports =2, name = 'Xf')
        self.Xr = wg.random(n_ports =2, name = 'Xr')
        self.Yf = wg.random(n_ports =2, name='Yf')
        self.Yr = wg.random(n_ports =2, name='Yr')
        self.If = wg.random(n_ports =1, name='If')
        self.Ir = wg.random(n_ports =1, name='Ir')

        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            None,
            ]
        actuals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.thru(),
            ]

        measured = [ self.measure(k) for k in actuals]

        self.cal = SOLT(
            ideals = ideals,
            measured = measured,
            n_thrus=1,
            isolation=measured[2]
            )


class TwoPortOnePathTest(TwelveTermTest):
    @suppress_warning_decorator("divide by zero encountered in log10")
    @suppress_warning_decorator("n_thrus is None")
    def setUp(self):
        self.n_ports = 2
        self.wg =WG
        wg  = self.wg
        self.Xf = wg.random(n_ports =2, name = 'Xf')
        self.Yf = wg.random(n_ports =2, name='Yf')

        #No leakage
        self.If = wg.match(n_ports =1, name='If')
        self.Ir = wg.match(n_ports =1, name='Ir')

        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.random(2,name='rand1'),
            wg.random(2,name='rand2'),
            ]

        measured = [ self.measure(k) for k in ideals]
        self.cal = TwoPortOnePath(
            ideals = ideals,
            measured = measured,
            source_port=1,
            #n_thrus=2,
            )

    def measure(self,ntwk):
        r = self.wg.match(2)
        m = ntwk.copy()
        mf = self.Xf**ntwk**self.Yf

        m.s[:,1,0] = mf.s[:,1,0]
        m.s[:,0,0] = mf.s[:,0,0]
        m.s[:,1,1] = r.s[:,1,1]
        m.s[:,0,1] = r.s[:,0,1]
        return m

    def test_accuracy_of_dut_correction(self):
        a = self.wg.random(n_ports=self.n_ports, name = 'actual')
        f = self.measure(a)
        r = self.measure(a.flipped())
        c = self.cal.apply_cal((f,r))
        c.name = 'corrected'
        self.assertEqual(c,a)

    def test_embed_then_apply_cal(self):
        a = self.wg.random(n_ports=self.n_ports)
        f = self.cal.embed(a)
        r = self.cal.embed(a.flipped())
        self.assertEqual(self.cal.apply_cal((f,r)),a)

    @pytest.mark.skip(reason='measurement procedure is different so this test doesnt apply')
    def test_embed_equal_measure(self):
        pass

    @suppress_warning_decorator("n_thrus is None")
    def test_from_coefs(self):
        cal_from_coefs = self.cal.from_coefs(self.cal.frequency, self.cal.coefs)
        ntwk = self.wg.random(n_ports=self.n_ports)

    @suppress_warning_decorator("n_thrus is None")
    def test_from_coefs_ntwks(self):
        cal_from_coefs = self.cal.from_coefs_ntwks(self.cal.coefs_ntwks)

    @pytest.mark.skip()
    def test_reverse_source_match_accuracy(self):
        pass

    @pytest.mark.skip()
    def test_reverse_directivity_accuracy(self):
        pass

    @pytest.mark.skip()
    def test_reverse_load_match_accuracy(self):
        pass

    @pytest.mark.skip()
    def test_reverse_reflection_tracking_accuracy(self):
        pass

    @pytest.mark.skip()
    def test_reverse_transmission_tracking_accuracy(self):
        pass

    @pytest.mark.skip()
    def test_convert_12term_2_8term_correction_accuracy(self):
        pass


class UnknownThruTest(EightTermTest):
    def setUp(self):

        self.n_ports = 2
        self.wg = WG
        wg= self.wg
        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        #No leakage
        self.If = wg.match(n_ports=1, name='If')
        self.Ir = wg.match(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')


        actuals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='match'),
            wg.impedance_mismatch(50,45)**wg.line(20,'deg',name='line')**wg.impedance_mismatch(45,50)

            ]

        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='match'),
            wg.thru(name='thru'),
            ]

        measured = [self.measure(k) for k in actuals]

        self.cal = rf.UnknownThru(
            ideals = ideals,
            measured = measured,
            switch_terms = [self.gamma_f, self.gamma_r]
            )

class LRMTest(EightTermTest):
    def setUp(self):

        self.n_ports = 2
        self.wg = WG
        wg = self.wg
        self.X = wg.random(n_ports=2, name='X')
        self.Y = wg.random(n_ports=2, name='Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports=1, name='gamma_f')
        self.gamma_r = wg.random(n_ports=1, name='gamma_r')

        # Our guess of the standards
        ss_i = wg.short(nports=2, name='short')
        thru = wg.line(d=50, unit='um', name='thru')

        # Actual reflects with parasitics
        s = wg.inductor(5e-12) ** wg.load(-0.95, nports=1, name='short')
        ss = rf.two_port_reflect(s, s)

        m = wg.load(0, nports=1, name='match')
        mm = rf.two_port_reflect(m, m)

        # Store reflect for other tests
        self.s = s

        approx_ideals = [
            thru,
            ss_i,
            mm
            ]

        ideals = [
            thru,
            ss,
            mm
            ]

        measured = [self.measure(k) for k in ideals]

        self.cal = rf.LRM(
            ideals = approx_ideals,
            measured = measured,
            switch_terms = [self.gamma_f, self.gamma_r],
            isolation = measured[2]
            )

    def test_solved_r(self):
        self.assertTrue(all(np.abs(self.s.s - self.cal.solved_r.s) < 1e-7))

class LRRMTest(EightTermTest):
    def setUp(self):

        self.n_ports = 2
        self.wg = WG
        wg = self.wg
        self.X = wg.random(n_ports=2, name='X')
        self.Y = wg.random(n_ports=2, name='Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports=1, name='gamma_f')
        self.gamma_r = wg.random(n_ports=1, name='gamma_r')

        # Our guess of the standards
        o_i = wg.load(1, nports=1, name='open')
        s_i = wg.short(nports=1, name='short')
        m_i = wg.load(0.1, nports=1, name='load')
        thru = wg.line(d=50, z0=75, unit='um', name='thru')
        thru.renormalize(wg.z0)

        # Make sure calibration works with non-symmetric thru
        thru.s[:,1,1] += 0.02 + 0.05j

        # Actual reflects with parasitics
        s = wg.inductor(5e-12) ** wg.load(-0.95, nports=1, name='short')
        o = wg.shunt_capacitor(5e-15) ** wg.open(nports=1, name='open')

        self.match_l = np.random.default_rng().uniform(1e-12, 20e-12)
        l = wg.inductor(L=self.match_l)
        m = l**m_i

        # Make standards two-ports
        oo = rf.two_port_reflect(o, o)
        ss = rf.two_port_reflect(s, s)
        mm = rf.two_port_reflect(m, o)

        oo_i = rf.two_port_reflect(o_i, o_i)
        ss_i = rf.two_port_reflect(s_i, s_i)
        mm_i = rf.two_port_reflect(m_i, m_i)

        # Store open and short for other tests.
        self.s = s
        self.o = o

        approx_ideals = [
            thru,
            ss_i,
            oo_i,
            mm_i
            ]

        ideals = [
            thru,
            ss,
            oo,
            mm
            ]

        measured = [self.measure(k) for k in ideals]

        self.cal = rf.LRRM(
            ideals = approx_ideals,
            measured = measured,
            switch_terms = [self.gamma_f, self.gamma_r],
            isolation = measured[3]
            )

    # Test the solved standards, don't use exact equality because of inductance
    # fitting tolerance.
    def test_solved_inductance(self):
        solved_l = np.mean(self.cal.solved_l)
        self.assertTrue(np.abs(self.match_l - solved_l) < 1e-3*self.match_l)

    def test_solved_r1(self):
        self.assertTrue(all(np.abs(self.s.s - self.cal.solved_r1.s) < 1e-7))

    def test_solved_r2(self):
        self.assertTrue(all(np.abs(self.o.s - self.cal.solved_r2.s) < 1e-7))

class LRRMTestNoFit(LRRMTest):
    def setUp(self):

        self.n_ports = 2
        self.wg = WG
        wg = self.wg
        self.X = wg.random(n_ports=2, name='X')
        self.Y = wg.random(n_ports=2, name='Y')
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports=1, name='gamma_f')
        self.gamma_r = wg.random(n_ports=1, name='gamma_r')

        # Our guess of the standards
        o_i = wg.load(1, nports=1, name='open')
        s_i = wg.short(nports=1, name='short')
        # Doesn't work correctly for non-50 ohm match.
        m_i = wg.load(0, nports=1, name='load')
        thru = wg.line(d=50, unit='um', name='thru')

        # Actual reflects with parasitics
        s = wg.inductor(5e-12) ** wg.load(-0.95, nports=1, name='short')
        o = wg.shunt_capacitor(5e-15) ** wg.open(nports=1, name='open')

        self.match_l = np.random.default_rng().uniform(1e-12, 20e-12)
        l = wg.inductor(L=self.match_l)
        m = l**m_i

        # Make standards two-ports
        oo = rf.two_port_reflect(o, o)
        ss = rf.two_port_reflect(s, s)
        mm = rf.two_port_reflect(m, o)

        oo_i = rf.two_port_reflect(o_i, o_i)
        ss_i = rf.two_port_reflect(s_i, s_i)
        mm_i = rf.two_port_reflect(m_i, m_i)

        # Store open and short for other tests.
        self.s = s
        self.o = o

        approx_ideals = [
            thru,
            ss_i,
            oo_i,
            mm_i
            ]

        ideals = [
            thru,
            ss,
            oo,
            mm
            ]

        measured = [self.measure(k) for k in ideals]

        self.cal = rf.LRRM(
            ideals = approx_ideals,
            measured = measured,
            switch_terms = [self.gamma_f, self.gamma_r],
            isolation = measured[3],
            match_fit = 'none'
            )

class MRCTest(EightTermTest):
    def setUp(self):

        self.n_ports = 2
        #Exact only with a lossless waveguide
        self.wg = WG_lossless
        wg= self.wg
        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        #No leakage
        self.If = wg.match(n_ports=1, name='If')
        self.Ir = wg.match(n_ports=1, name='Ir')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')


        def delay_shorts(d1,d2):
            ds1 = wg.delay_short(d1,'deg')
            ds2 = wg.delay_short(d2,'deg')
            return rf.two_port_reflect(ds1,ds2)

        actuals = [
            wg.short(nports=2, name='short'),
            delay_shorts(65,130),
            delay_shorts(120,75),
            wg.load(.2+.2j,nports=2, name='match'),
            wg.impedance_mismatch(50,45)**wg.line(20,'deg',name='line')**wg.impedance_mismatch(45,50)

            ]

        ideals = [
            wg.short(nports=2, name='short'),
            delay_shorts(45,90),
            delay_shorts(90,45),
            wg.load(.2+.2j,nports=2, name='match'),
            wg.thru(name='thru'),
            ]

        measured = [self.measure(k) for k in actuals]

        self.cal = rf.MRC(
            ideals = ideals,
            measured = measured,
            switch_terms = [self.gamma_f, self.gamma_r]
            )


class TwelveTermToEightTermTest(unittest.TestCase, CalibrationTest):
    """
    This test verifies the accuracy of the SOLT calibration, when used
    on an error-box (8-term) model.


    """
    @suppress_warning_decorator("n_thrus is None")
    def setUp(self):
        self.n_ports = 2
        wg= rf.wr10
        wg.frequency = rf.F.from_f([100], unit='GHz')
        self.wg = wg
        self.X = wg.random(n_ports =2, name = 'X')
        self.Y = wg.random(n_ports =2, name='Y')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')
        #Isolation terms
        self.If = wg.random(n_ports=1, name='If')
        self.Ir = wg.random(n_ports=1, name='Ir')

        ideals = [
            wg.short(nports=2, name='short'),
            wg.open(nports=2, name='open'),
            wg.match(nports=2, name='load'),
            wg.thru(name='thru'),
            ]

        measured = [ self.measure(k) for k in ideals]
        with warnings.catch_warnings(record=False):
            self.cal = rf.TwelveTerm(
            ideals = ideals,
            measured = measured,
            isolation=measured[2]
            )


        coefs = rf.calibration.convert_12term_2_8term(self.cal.coefs)
        coefs = NetworkSet.from_s_dict(coefs,
                                    frequency=self.cal.frequency).to_dict()
        self.coefs= coefs

    def terminate(self, ntwk):
        """
        terminate a measured network with the switch terms
        """
        return terminate(ntwk,self.gamma_f, self.gamma_r)

    def measure(self,ntwk):
        out =  self.terminate(self.X**ntwk**self.Y)
        out.name = ntwk.name
        out.s[:,1,0] += self.If.s[:,0,0]
        out.s[:,0,1] += self.Ir.s[:,0,0]
        return out

    def test_forward_isolation(self):
        self.assertEqual(self.coefs['forward isolation'], self.If.s11)

    def test_reverse_isolation(self):
        self.assertEqual(self.coefs['reverse isolation'], self.Ir.s11)

    def test_forward_switch_term(self):
        self.assertEqual(self.coefs['forward switch term'], self.gamma_f)

    def test_reverse_switch_term(self):
        self.assertEqual(self.coefs['reverse switch term'], self.gamma_r)

    def test_forward_directivity_accuracy(self):
        self.assertEqual(
            self.X.s11,
            self.coefs['forward directivity'])

    def test_forward_source_match_accuracy(self):
        self.assertEqual(
            self.X.s22 ,
            self.coefs['forward source match'] )

    def test_forward_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.X.s21 * self.X.s12 ,
            self.coefs['forward reflection tracking'])

    def test_reverse_source_match_accuracy(self):
        self.assertEqual(
            self.Y.s11 ,
            self.coefs['reverse source match']   )

    def test_reverse_directivity_accuracy(self):
        self.assertEqual(
            self.Y.s22 ,
            self.coefs['reverse directivity']  )

    def test_reverse_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.Y.s21 * self.Y.s12 ,
            self.coefs['reverse reflection tracking'])

    def test_k_accuracy(self):
        self.assertEqual(
            self.X.s21/self.Y.s12 ,
            self.coefs['k']  )

    def test_verify_12term(self):
        self.assertTrue(self.cal.verify_12term_ntwk.s_mag.max() < 1e-3)


class SixteenTermTest(unittest.TestCase, CalibrationTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg = self.wg

        #Port 0: VNA port 0
        #Port 1: DUT port 0
        #Port 2: DUT port 1
        #Port 3: VNA port 1
        self.Z = wg.random(n_ports = 4, name = 'Z')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        o = wg.open(nports=1, name='open')
        s = wg.short(nports=1, name='short')
        m = wg.match(nports=1, name='load')
        om = rf.two_port_reflect(o, m)
        mo = rf.two_port_reflect(m, o)
        oo = rf.two_port_reflect(o, o)
        ss = rf.two_port_reflect(s, s)
        thru = wg.thru(name='thru')

        ideals = [
            thru,
            om,
            mo,
            oo,
            ss
            ]

        measured = [self.measure(k) for k in ideals]

        self.cal = rf.SixteenTerm(
            measured = measured,
            ideals = ideals,
            switch_terms=(self.gamma_f, self.gamma_r)
            )

    def terminate(self, ntwk):
        """
        terminate a measured network with the switch terms
        """
        return terminate(ntwk,self.gamma_f, self.gamma_r)

    def measure(self,ntwk):
        out = self.terminate(rf.connect(self.Z, 1, ntwk, 0, num=2))
        out.name = ntwk.name
        return out

    def test_forward_directivity_accuracy(self):
        self.assertEqual(
            self.Z.s11,
            self.cal.coefs_ntwks['forward directivity'])

    def test_forward_source_match_accuracy(self):
        self.assertEqual(
            self.Z.s22 ,
            self.cal.coefs_ntwks['forward source match'] )

    def test_forward_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.Z.s21 * self.Z.s12 ,
            self.cal.coefs_ntwks['forward reflection tracking'])

    def test_reverse_source_match_accuracy(self):
        self.assertEqual(
            self.Z.s33 ,
            self.cal.coefs_ntwks['reverse source match']   )

    def test_reverse_directivity_accuracy(self):
        self.assertEqual(
            self.Z.s44 ,
            self.cal.coefs_ntwks['reverse directivity']  )

    def test_reverse_reflection_tracking_accuracy(self):
        self.assertEqual(
            self.Z.s34 * self.Z.s43 ,
            self.cal.coefs_ntwks['reverse reflection tracking'])

    def test_forward_isolation(self):
        self.assertEqual(
            self.Z.s41 ,
            self.cal.coefs_ntwks['forward isolation'])

    def test_reverse_isolation(self):
        self.assertEqual(
            self.Z.s14 ,
            self.cal.coefs_ntwks['reverse isolation'])

    def test_reverse_port_isolation(self):
        self.assertEqual(
            self.Z.s23 ,
            self.cal.coefs_ntwks['reverse port isolation'])

    def test_forward_port_isolation(self):
        self.assertEqual(
            self.Z.s32 ,
            self.cal.coefs_ntwks['forward port isolation'])

    def test_k(self):
        self.assertEqual(
            self.Z.s21/self.Z.s34 ,
            self.cal.coefs_ntwks['k'])

    def test_forward_port1_isolation(self):
        self.assertEqual(
            self.Z.s31/self.Z.s34 ,
            self.cal.coefs_ntwks['forward port 1 isolation'])

    def test_reverse_port1_isolation(self):
        self.assertEqual(
            self.Z.s13*self.Z.s34 ,
            self.cal.coefs_ntwks['reverse port 1 isolation'])

    def test_forward_port2_isolation(self):
        self.assertEqual(
            self.Z.s42*self.Z.s34 ,
            self.cal.coefs_ntwks['forward port 2 isolation'])

    def test_reverse_port2_isolation(self):
        self.assertEqual(
            self.Z.s24/self.Z.s34 ,
            self.cal.coefs_ntwks['reverse port 2 isolation'])

class SixteenTermCoefficientsTest(unittest.TestCase):
    """Test that 16-term non-isolation coefficients are defined the same way as
    8-term coefficients."""
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg = self.wg

        #Port 0: VNA port 0
        #Port 1: DUT port 0
        #Port 2: DUT port 1
        #Port 3: VNA port 1
        self.Z = wg.random(n_ports = 4, name = 'Z')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        o = wg.open(nports=1, name='open')
        s = wg.short(nports=1, name='short')
        m = wg.match(nports=1, name='load')
        om = rf.two_port_reflect(o, m)
        mo = rf.two_port_reflect(m, o)
        oo = rf.two_port_reflect(o, o)
        ss = rf.two_port_reflect(s, s)
        thru = wg.thru(name='thru')

        ideals = [
            thru,
            om,
            mo,
            oo,
            ss
            ]

        measured = [self.measure(k) for k in ideals]

        self.cal16 = rf.SixteenTerm(
            measured = measured,
            ideals = ideals,
            switch_terms=(self.gamma_f, self.gamma_r)
            )

        r = wg.load(.95+.1j,nports=1)
        m = wg.match(nports=1)
        mm = rf.two_port_reflect(m, m)
        rm = rf.two_port_reflect(r, m)
        mr = rf.two_port_reflect(m, r)
        rr = rf.two_port_reflect(r, r)

        ideals = [
            thru,
            mm,
            rr,
            rm,
            mr
            ]

        measured = [self.measure(k) for k in ideals]

        self.cal_lmr16 = rf.LMR16(
            measured = measured,
            ideals = [thru],
            ideal_is_reflect = False,
             #Automatic sign detection doesn't work if the
             #error terms aren't symmetric enough
            sign = 1,
            switch_terms=(self.gamma_f, self.gamma_r)
            )

        #Same error network, but without leakage terms

        #Primary leakage
        self.Z.s[:,3,0] = 0 # forward isolation
        self.Z.s[:,0,3] = 0 # reverse isolation
        self.Z.s[:,2,1] = 0 # forward port isolation
        self.Z.s[:,1,2] = 0 # reverse port isolation

        #Cross leakage
        self.Z.s[:,3,1] = 0 # forward port 2 isolation
        self.Z.s[:,1,3] = 0 # reverse port 2 isolation
        self.Z.s[:,2,0] = 0 # forward port 1 isolation
        self.Z.s[:,0,2] = 0 # reverse port 1 isolation

        measured = [self.measure(k) for k in ideals]

        self.cal8 = rf.EightTerm(
            measured = measured,
            ideals = ideals,
            switch_terms=(self.gamma_f, self.gamma_r),
            )

    def terminate(self, ntwk):
        """
        terminate a measured network with the switch terms
        """
        return terminate(ntwk,self.gamma_f, self.gamma_r)

    def measure(self,ntwk):
        out = self.terminate(rf.connect(self.Z, 1, ntwk, 0, num=2))
        out.name = ntwk.name
        return out

    def test_coefficients(self):
        for k in self.cal8.coefs.keys():
            if k in self.cal16.coefs.keys():
                if 'isolation' in k:
                    continue
                self.assertTrue(all(np.abs(self.cal8.coefs[k] - self.cal16.coefs[k]) < 1e-10))
                self.assertTrue(all(np.abs(self.cal8.coefs[k] - self.cal_lmr16.coefs[k]) < 1e-10))


class LMR16Test(SixteenTermTest):
    def setUp(self):
        self.n_ports = 2
        self.wg = WG
        wg = self.wg

        #Port 0: VNA port 0
        #Port 1: DUT port 0
        #Port 2: DUT port 1
        #Port 3: VNA port 1
        self.Z = wg.random(n_ports = 4, name = 'Z')
        self.gamma_f = wg.random(n_ports =1, name='gamma_f')
        self.gamma_r = wg.random(n_ports =1, name='gamma_r')

        r = wg.short(nports=1, name='short')
        m = wg.match(nports=1, name='load')
        mm = rf.two_port_reflect(m, m)
        rm = rf.two_port_reflect(r, m)
        mr = rf.two_port_reflect(m, r)
        rr = rf.two_port_reflect(r, r)

        thru_length = np.random.default_rng().uniform(0,10)
        thru = wg.line(thru_length,'deg',name='line')

        self.thru = thru

        ideals = [
            thru,
            mm,
            rr,
            rm,
            mr
            ]

        measured = [self.measure(k) for k in ideals]

        self.cal = rf.LMR16(
            measured = measured,
            ideals = [r],
            ideal_is_reflect = True,
             #Automatic sign detection doesn't work if the
             #error terms aren't symmetric enough
            sign = -1,
            switch_terms=(self.gamma_f, self.gamma_r)
            )

    def test_solved_through(self):
        self.assertEqual(
            self.thru,
            self.cal.solved_through)

class MultiportCalTest(unittest.TestCase):
    """Multi-port NISTMultilineTRL calibration test"""

    def test_cal(self):
        self.wg = WG
        wg = self.wg

        self.n_ports = 3
        nports = self.n_ports

        self.make_error_networks(wg, nports)

        cal_dict = {}

        actuals = [
            wg.thru(),
            rf.two_port_reflect(wg.load(-.98-.1j),wg.load(-.98-.1j)),
            wg.line(100,'um'),
            wg.line(900,'um'),
            ]

        dut = self.wg.random(n_ports=self.n_ports, name='dut')
        dut_meas = self.measure(dut)

        l = [0, 100e-6, 900e-6]

        for p in [(0, 1), (0, 2)]:

            measured = [self.measure(rf.twoport_to_nport(k, p[0], p[1], nports)) for k in actuals]
            cal_dict[p] = {}
            cal_dict[p]['method'] = NISTMultilineTRL
            cal_dict[p]['measured'] = measured
            # Two-port switch terms are in reverse order from multi-port switch terms.
            cal_dict[p]['switch_terms'] = [self.gammas[i] for i in p][::-1]
            cal_dict[p]['l'] = l
            cal_dict[p]['Grefls'] = [-1]
            cal_dict[p]['er_est'] = 1
            cal_dict[p]['gamma_root_choice'] = 'real'

        isolation = self.measure(wg.match(nports=nports))

        self.cal = MultiportCal(cal_dict, isolation=isolation)

        # Test coefs
        nports = self.n_ports
        for e, c in enumerate(self.cal.coefs):
            assert np.allclose(c['directivity'], self.Z.s[:,e,e])
            assert np.allclose(c['source match'], self.Z.s[:,nports+e,nports+e])
            assert np.allclose(c['reflection tracking'], self.Z.s[:,e,nports+e] * self.Z.s[:,nports+e,e])
            assert np.allclose(c['switch term'], self.gammas[e].s)
            assert np.allclose(c['k']/self.cal.coefs[0]['k'], self.Z.s[:,nports+0,0]/self.Z.s[:,nports+e,e])

        # Test DUT correction
        dut_cal = self.cal.apply_cal(dut_meas)
        self.assertEqual(dut_cal, dut)

        # Test cal inverts embed
        a = self.wg.random(n_ports=self.n_ports)
        self.assertEqual(self.cal.apply_cal(self.cal.embed(a)),a)

        # Test embed equals measure
        a = self.wg.random(n_ports=self.n_ports)
        self.assertEqual(self.cal.embed(a),self.measure(a))

        # Test gamma solved by TRL
        for p in [(0, 1), (0, 2)]:
            self.assertTrue(max(np.abs(self.wg.gamma-self.cal.cals[p].gamma)) < 1e-3)

    def make_error_networks(self, wg, nports):
        self.Z = wg.random(n_ports = 2*nports, name = 'Z')

        # Isolation terms are between all ports.
        # No error in the same port.
        self.isolation = wg.random(n_ports=nports, name='I')
        for i in range(nports):
            self.isolation.s[:, i, i] = 0

        def port_type(n):
            return 'VNA' if n < nports else 'DUT'
        def port_number(n):
            return n if n < nports else n - nports
        # Remove leakage terms
        for i in range(2*nports):
            for j in range(i+1, 2*nports):
                # No connection between different VNA/DUT ports.
                # No connection between VNA and DUT ports with different number.
                if port_type(i) == port_type(j) or port_number(i) != port_number(j):
                    self.Z.s[:,i,j] = 0
                    self.Z.s[:,j,i] = 0

        self.gammas = []
        for i in range(nports):
            self.gammas.append(wg.random(n_ports=1, name=f'gamma_{i}'))

    def terminate(self, ntwk):
        """
        Terminate a measured network with the switch terms
        """
        return terminate_nport(ntwk, self.gammas)

    def measure(self, ntwk):
        out = self.terminate(rf.connect(self.Z, self.n_ports, ntwk, 0, num=self.n_ports))
        out = out + self.isolation
        out.name = ntwk.name
        return out

class MultiportSOLTTest(MultiportCalTest):

    def test_cal(self):
        nport_list = [3, 4]
        method_list = [SOLT, UnknownThru]

        self.wg = WG
        wg = self.wg

        for nport in nport_list:
            self.n_ports = nport
            nports = self.n_ports

            self.make_error_networks(wg, nports)

            o = wg.open(nports=nports, name='open')
            s = wg.short(nports=nports, name='short')
            m = wg.match(nports=nports, name='load')

            # thru = wg.thru(name='thru')
            thru = wg.impedance_mismatch(50, 45) ** wg.line(20, 'deg') ** wg.impedance_mismatch(45, 50)

            ideals = []
            # nports-1 thrus from port 0 to all other ports.
            for i in range(1, nports):
                thru_i = rf.twoport_to_nport(thru, 0, i, nports)
                ideals.append(thru_i)

            ideals.extend([o,s,m])
            measured = [self.measure(k) for k in ideals]

            dut = self.wg.random(n_ports=self.n_ports, name='dut')
            dut_meas = self.measure(dut)

            for method in method_list:

                if method == SOLT:
                    self.cal = MultiportSOLT(method, measured, ideals, isolation=measured[-1])
                else:
                    self.cal = MultiportSOLT(method, measured, ideals, isolation=measured[-1], switch_terms=self.gammas)

                # Test coefs
                nports = self.n_ports
                for e, c in enumerate(self.cal.coefs):
                    assert np.allclose(c['directivity'], self.Z.s[:,e,e])
                    assert np.allclose(c['source match'], self.Z.s[:,nports+e,nports+e])
                    assert np.allclose(c['reflection tracking'], self.Z.s[:,e,nports+e] * self.Z.s[:,nports+e,e])
                    assert np.allclose(c['switch term'], self.gammas[e].s)
                    assert np.allclose(c['k']/self.cal.coefs[0]['k'], self.Z.s[:,nports+0,0]/self.Z.s[:,nports+e,e])

                # Test DUT correction
                dut_cal = self.cal.apply_cal(dut_meas)
                self.assertEqual(dut_cal, dut)

                # Test cal inverts embed
                a = self.wg.random(n_ports=self.n_ports)
                self.assertEqual(self.cal.apply_cal(self.cal.embed(a)),a)

                # Test embed equals measure
                a = self.wg.random(n_ports=self.n_ports)
                self.assertEqual(self.cal.embed(a),self.measure(a))

if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Sequence

import itertools
import sys
from enum import Enum

import numpy as np

import skrf
from skrf.vi import vna
from skrf.vi.validators import (
    BooleanValidator,
    DelimitedStrValidator,
    EnumValidator,
    FloatValidator,
    FreqValidator,
    IntValidator,
)
from skrf.vi.vna import VNA, ValuesFormat


class SweepType(Enum):
    LINEAR = "LIN"
    LOG = "LOG"
    POWER = "POW"
    CW = "CW"
    SEGMENT = "SEGM"
    PHASE = "PHAS"


class SweepMode(Enum):
    HOLD = "HOLD"
    CONTINUOUS = "CONT"
    GROUPS = "GRO"
    SINGLE = "SING"


class TriggerSource(Enum):
    EXTERNAL = "EXT"
    IMMEDIATE = "IMM"
    MANUAL = "MAN"


class AveragingMode(Enum):
    POINT = "POIN"
    SWEEP = "SWE"


class PNA(VNA):
    '''
    Keysight PNA, PNA-L, PNA-X.

    PNA Models
    ==========
    E8361A, E8362B, E8363B, E8364B, N5221A, N5222A, N5224A, N5225A, N5227A
    N5224B, N5222B, N5227B, N5225B, N5221B, E8356A, E8357A, E8358A, E8361A
    E8361C, E8362A, E8362C, E8363A, E8363B, E8363C, E8364A, E8364B, E8364C
    E8801A, E8802A, E8803A, N3381A, N3382A, N3383A, N5250C

    PNA-L Models
    ============
    N5230A, N5230C, N5231A, N5232A, N5234A, N5235A, N5239A, N5234B, N5235B
    N5231B, N5232B, N5239B

    PNA-X Models
    ============
    N5241A, N5242A, N5244A, N5245A, N5247A, N5249A, N5247B, N5245B, N5244B
    N5242B, N5241B, N5249B, N5264A, N5264B
    '''
    # TODO:
    # - active_calset_name: SENS<self:cnum>:CORR:CSET:ACT? NAME
    # - create_calset: SENS<self:cnum>:CORR:CSET:CRE <name>
    # - calset_data: SENS<self:cnum>:CORR:CSET:DATA?  <eterm>,<port a>,<port b> '<receiver>'


    _models = {
        "default": {"nports": 2, "unsupported": []},
        "E8362C": {"nports": 2, "unsupported": ["nports", "freq_step", "fast_sweep"]},
        "N5227B": {"nports": 4, "unsupported": []},
    }

    class Channel(vna.Channel):
        def __init__(self, parent, cnum: int, cname: str):
            super().__init__(parent, cnum, cname)

            if cnum != 1:
                default_msmnt = f"CH{self.cnum}_S11_1"
                self.create_measurement(default_msmnt, "S11")

        def _on_delete(self):
            self.write(f"SYST:CHAN:DEL {self.cnum}")

        freq_start = VNA.command(
            get_cmd="SENS<self:cnum>:FREQ:STAR?",
            set_cmd="SENS<self:cnum>:FREQ:STAR <arg>",
            doc="""The start frequency [Hz]""",
            validator=FreqValidator(),
        )

        freq_stop = VNA.command(
            get_cmd="SENS<self:cnum>:FREQ:STOP?",
            set_cmd="SENS<self:cnum>:FREQ:STOP <arg>",
            doc="""The stop frequency [Hz]""",
            validator=FreqValidator(),
        )

        freq_span = VNA.command(
            get_cmd="SENS<self:cnum>:FREQ:SPAN?",
            set_cmd="SENS<self:cnum>:FREQ:SPAN <arg>",
            doc="""The frequency span [Hz].""",
            validator=FreqValidator(),
        )

        freq_center = VNA.command(
            get_cmd="SENS<self:cnum>:FREQ:CENT?",
            set_cmd="SENS<self:cnum>:FREQ:CENT <arg>",
            doc="""The frequency span [Hz].""",
            validator=FreqValidator(),
        )

        npoints = VNA.command(
            get_cmd="SENS<self:cnum>:SWE:POIN?",
            set_cmd="SENS<self:cnum>:SWE:POIN <arg>",
            doc="""The number of frequency points. Sets the frequency step as a
                side effect
            """,
            validator=IntValidator(),
        )

        if_bandwidth = VNA.command(
            get_cmd="SENS<self:cnum>:BWID?",
            set_cmd="SENS<self:cnum>:BWID <arg>",
            doc="""The IF bandwidth [Hz]""",
            validator=FreqValidator(),
        )

        sweep_time = VNA.command(
            get_cmd="SENS<self:cnum>:SWE:TIME?",
            set_cmd="SENS<self:cnum>:SWE:TIME <arg>",
            doc="""The time in seconds for a single sweep [s]""",
            validator=FloatValidator(decimal_places=2),
        )

        sweep_type = VNA.command(
            get_cmd="SENS<self:cnum>:SWE:TYPE?",
            set_cmd="SENS<self:cnum>:SWE:TYPE <arg>",
            doc="""The type of sweep (linear, log, etc)""",
            validator=EnumValidator(SweepType),
        )

        sweep_mode = VNA.command(
            get_cmd="SENS<self:cnum>:SWE:MODE?",
            set_cmd="SENS<self:cnum>:SWE:MODE <arg>",
            doc="""This channel's trigger mode""",
            validator=EnumValidator(SweepMode),
        )

        measurement_numbers = VNA.command(
            get_cmd="SYST:MEAS:CAT? <self:cnum>",
            set_cmd=None,
            doc="""The list of measurement numbers on this channel""",
            validator=DelimitedStrValidator(int),
        )

        averaging_on = VNA.command(
            get_cmd="SENS<self:cnum>:AVER:STATE?",
            set_cmd="SENS<self:cnum>:AVER:STATE <arg>",
            doc="""Whether averaging is on or off""",
            validator=BooleanValidator(),
        )

        averaging_count = VNA.command(
            get_cmd="SENS<self:cnum>:AVER:COUN?",
            set_cmd="SENS<self:cnum>:AVER:COUN <arg>",
            doc="""The number of measurements combined for an average""",
            validator=IntValidator(1, 65536),
        )

        averaging_mode = VNA.command(
            get_cmd="SENS<self:cnum>:AVER:MODE?",
            set_cmd="SENS<self:cnum>:AVER:MODE <arg>",
            doc="""How measurements are averaged together""",
            validator=EnumValidator(AveragingMode),
        )

        n_sweep_groups = VNA.command(
            get_cmd="SENS<self:cnum>:SWE:GRO:COUN?",
            set_cmd="SENS<self:cnum>:SWE:GRO:COUN <arg>",
            doc="""The number of triggers sent for one trigger command""",
            validator=IntValidator(1, int(2e6)),
        )

        @property
        def freq_step(self) -> int:
            # Not all instruments support SENS:FREQ:STEP
            f = self.frequency
            return int(f.step)

        @freq_step.setter
        def freq_step(self, f: int | float | str) -> None:
            validator = FreqValidator()
            f = validator.validate_input(f)
            freq = self.frequency
            self.npoints = len(range(int(freq.start), int(freq.stop) + f, f))

        @property
        def frequency(self) -> skrf.Frequency:
            f = skrf.Frequency(
                start=self.freq_start,
                stop=self.freq_stop,
                npoints=self.npoints,
                unit="hz",
            )
            return f

        @frequency.setter
        def frequency(self, f: skrf.Frequency) -> None:
            self.freq_start = f.start
            self.freq_stop = f.stop
            self.npoints = f.npoints

        @property
        def measurements(self) -> list[tuple[str, str]]:
            msmnts = self.query(f"CALC{self.cnum}:PAR:CAT:EXT?").replace('"', "")
            msmnts = msmnts.split(",")
            return list(zip(msmnts[::2], msmnts[1::2]))

        @property
        def measurement_names(self) -> list[str]:
            return [msmnt[0] for msmnt in self.measurements]

        @property
        def calibration(self) -> skrf.Calibration:
            raise NotImplementedError()
            # orig_query_fmt = self.parent.query_format
            # self.parent.query_format = ValuesFormat.BINARY_64

            # eterm_re = re.compile(r"(\w+)\((\d),(\d)\)")
            # cset_terms = self.query(f"SENS{self.cnum}:CORR:CSET:ETER:CAT?")
            # terms = re.findall(eterm_re, cset_terms)

            # if len(terms) == 3:
            #     cal = skrf.OnePort
            # elif len(terms) == 12:
            #     # cal = skrf.TwoPort
            #     raise NotImplementedError()
            # else:
            #     # cal = skrf.MultiPort
            #     raise NotImplementedError()

            # coeffs = {}
            # for term in terms:
            #     pna_name = term[0]
            #     skrf_name = re.sub(
            #         r"([A-Z])", lambda pat: f" {pat.group(1).lower()}", pna_name
            #     ).strip()

            #     coeffs[skrf_name] = self.query_values(
            #         f"SENS{self.cnum}:CORR:CSET:ETERM? '{pna_name}({a},{b})'",
            #         complex_values=True,
            #         container=np.array,
            #     )

            # self.parent.query_format = orig_query_fmt

            # freq = self.frequency
            # return cal.from_coefs(freq, coeffs)

        @calibration.setter
        def calibration(self, cal: skrf.Calibration) -> None:
            raise NotImplementedError()

        @property
        def active_trace_sdata(self) -> np.ndarray:
            active_measurement = (
                self.query(f"CALC{self.cnum}:PAR:SEL?").replace('"', "").split(",")[0]
            )
            if active_measurement == "":
                raise RuntimeError("No trace is active. Must select measurement first.")
            return self.query_values(
                f"CALC{self.cnum}:DATA? SDATA", complex_values=True
            )

        def clear_averaging(self) -> None:
            self.write(f"SENS{self.cnum}:AVER:CLE")

        def create_measurement(self, name: str, parameter: str) -> None:
            self.write(f"CALC{self.cnum}:PAR:EXT '{name}',{parameter}")
            # Not all instruments support DISP:WIND:TRAC:NEXT
            traces = self.query("DISP:WIND:CAT?").replace('"', "")
            traces = [int(tr) for tr in traces.split(",")] if traces != "EMPTY" else [0]
            next_tr = traces[-1] + 1
            self.write(f"DISP:WIND:TRAC{next_tr}:FEED '{name}'")

        def delete_measurement(self, name: str) -> None:
            self.write(f"CALC{self.cnum}:PAR:DEL '{name}'")

        def get_measurement(self, name: str) -> skrf.Network:
            if name not in self.measurement_names:
                raise KeyError(f"{name} does not exist")

            self.parent.active_measurement = name
            ntwk = self.get_active_trace()
            ntwk.name = name
            return ntwk

        def get_active_trace(self) -> skrf.Network:
            self.sweep()
            orig_query_fmt = self.parent.query_format
            self.parent.query_format = ValuesFormat.BINARY_64

            ntwk = skrf.Network()
            ntwk.frequency = self.frequency
            ntwk.s = self.active_trace_sdata

            self.parent.query_format = orig_query_fmt

            return ntwk

        def get_sdata(self, a: int | str, b: int | str) -> skrf.Network:
            self.sweep()
            orig_query_fmt = self.parent.query_format
            self.parent.query_format = ValuesFormat.BINARY_64
            param = f"S{a}{b}"
            self.create_measurement("SKRF_TMP", param)
            self.parent.active_measurement = "SKRF_TMP"

            ntwk = skrf.Network()
            ntwk.frequency = self.frequency
            ntwk.s = self.active_trace_sdata

            self.delete_measurement("SKRF_TMP")
            self.parent.query_format = orig_query_fmt

            return ntwk

        def get_snp_network(
            self,
            ports: Sequence | None = None,
        ) -> skrf.Network:
            if ports is None:
                ports = list(range(1, self.parent.nports + 1))

            orig_query_fmt = self.parent.query_format
            self.parent.query_format = ValuesFormat.BINARY_64
            self.parent.active_channel = self
            orig_snp_fmt = self.query("MMEM:STOR:TRAC:FORM:SNP?")
            self.write("MMEM:STOR:TRACE:FORM:SNP RI") # Expect Real/Imaginary data

            msmnt_params = [f"S{a}{b}" for a, b in itertools.product(ports, repeat=2)]

            names = []
            # Make sure the ports specified are driven
            for param in msmnt_params:
                # Not all models support CALC:PAR:TAG:NEXT
                name = f"CH{self.cnum}_SKRF_{param}"
                names.append(name)
                self.create_measurement(name, param)

            self.sweep()
            port_str = ",".join(str(port) for port in ports)
            raw = self.query_values(
                f"CALC{self.cnum}:DATA:SNP:PORTS? '{port_str}'", container=np.array
            )
            self.parent.wait_for_complete()

            for name in names:
                self.delete_measurement(name)

            # The data is sent back as:
            # [
            #   [frequency points],
            #   [s11.real],
            #   [s11.imag],
            #   [s12.real],
            #   [s12.imag],
            # ...
            # ]
            # but flattened. So we recreate the above shape from the flattened data
            npoints = self.npoints
            nrows = len(raw) // npoints
            nports = len(ports)
            data = raw.reshape((nrows, -1))[1:]

            ntwk = skrf.Network()
            ntwk.frequency = self.frequency
            ntwk.s = np.empty(
                shape=(len(ntwk.frequency), nports, nports), dtype=complex
            )
            real_rows = data[::2]
            imag_rows = data[1::2]
            for n in range(nports):
                for m in range(nports):
                    i = n * nports + m
                    ntwk.s[:, n, m] = real_rows[i] + 1j * imag_rows[i]

            self.parent.query_format = orig_query_fmt
            self.write(f"MMEM:STOR:TRACE:FORM:SNP {orig_snp_fmt}")

            return ntwk

        def sweep(self) -> None:
            self.parent.trigger_source = TriggerSource.IMMEDIATE
            self.parent._resource.clear()

            sweep_mode = self.sweep_mode
            sweep_time = self.sweep_time
            avg_on = self.averaging_on
            avg_mode = self.averaging_mode

            original_config = {
                "sweep_mode": sweep_mode,
                "sweep_time": sweep_time,
                "averaging_on": avg_on,
                "averaging_mode": avg_mode,
                "timeout": self.parent._resource.timeout,
            }

            if avg_on and avg_mode == AveragingMode.SWEEP:
                self.sweep_mode = SweepMode.GROUPS
                n_sweeps = self.averaging_count
                self.n_sweep_groups = n_sweeps
                n_sweeps *= self.parent.nports
            else:
                self.sweep_mode = SweepMode.SINGLE
                n_sweeps = self.parent.nports

            try:
                sweep_time *= n_sweeps * 1_000  # 1s per port
                self.parent._resource.timeout = max(sweep_time, 5_000)  # minimum of 5s
                self.parent.wait_for_complete()
            finally:
                self.parent._resource.clear()
                self.parent._resource.timeout = original_config.pop("timeout")
                for k, v in original_config.items():
                    setattr(self, k, v)

    def __init__(self, address: str, backend: str = "@py") -> None:
        super().__init__(address, backend)

        self._resource.read_termination = "\n"
        self._resource.write_termination = "\n"

        self.create_channel(1, "Channel 1")
        self.active_channel = self.ch1

        self.model = self.id.split(",")[1]
        if self.model not in self._models:
            print(
                f"WARNING: This model ({self.model}) has not been tested with "
                "scikit-rf. By default, all features are turned on but older "
                "instruments might be missing SCPI support for some commands "
                "which will cause errors. Consider submitting an issue on GitHub to "
                "help testing and adding support.",
                file=sys.stderr,
            )

    def _supports(self, feature: str) -> bool:
        model_config = self._models.get(self.model, self._models["default"])
        return feature not in model_config["unsupported"]

    def _model_param(self, param: str):
        model_config = self._models.get(self.model, self._models["default"])
        return model_config[param]

    trigger_source = VNA.command(
        get_cmd="TRIG:SOUR?",
        set_cmd="TRIG:SOUR <arg>",
        doc="""The source of the sweep trigger signal""",
        validator=EnumValidator(TriggerSource),
    )

    nerrors = VNA.command(
        get_cmd="SYST:ERR:COUN?",
        set_cmd=None,
        doc="""The number of errors since last cleared (see
            :func:`PNA.clear_errors`)""",
        validator=IntValidator(),
    )

    channel_numbers = VNA.command(
        get_cmd="SYST:CHAN:CAT?",
        set_cmd=None,
        doc="""The channel numbers currently in use""",
        validator=DelimitedStrValidator(int),
    )

    @property
    def nports(self) -> int:
        if self._supports("nports"):
            return int(self.query("SYST:CAP:HARD:PORT:COUN?"))
        else:
            return self._model_param("nports")

    @property
    def active_channel(self) -> Channel | None:
        num = int(self.query("SYST:ACT:CHAN?"))
        return getattr(self, f"ch{num}", None)

    @active_channel.setter
    def active_channel(self, ch: Channel) -> None:
        if self.active_channel.cnum == ch.cnum:
            return

        msmnt = ch.measurement_numbers[0]
        self.write(f"CALC{ch.cnum}:PAR:MNUM {msmnt}")

    @property
    def query_format(self) -> vna.ValuesFormat:
        fmt = self.query("FORM?").replace("+", "")
        if fmt == "ASC,0":
            self._values_fmt = vna.ValuesFormat.ASCII
        elif fmt == "REAL,32":
            self._values_fmt = vna.ValuesFormat.BINARY_32
        elif fmt == "REAL,64":
            self._values_fmt = vna.ValuesFormat.BINARY_64
        return self._values_fmt

    @query_format.setter
    def query_format(self, fmt: vna.ValuesFormat) -> None:
        if fmt == vna.ValuesFormat.ASCII:
            self._values_fmt = vna.ValuesFormat.ASCII
            self.write("FORM ASC,0")
        elif fmt == vna.ValuesFormat.BINARY_32:
            self._values_fmt = vna.ValuesFormat.BINARY_32
            self.write("FORM:BORD SWAP")
            self.write("FORM REAL,32")
        elif fmt == vna.ValuesFormat.BINARY_64:
            self._values_fmt = vna.ValuesFormat.BINARY_64
            self.write("FORM:BORD SWAP")
            self.write("FORM REAL,64")

    @property
    def active_measurement(self) -> str:
        return self.query("SYST:ACT:MEAS?").replace('"', "")

    @active_measurement.setter
    def active_measurement(self, name: str) -> None:
        measurements = {
            name: channel
            for channel in self.channels
            for name in channel.measurement_names
        }

        if name not in measurements:
            raise KeyError(f"{name} does not exist")

        if self._supports("fast_sweep"):
            self.write(f"CALC{measurements[name].cnum}:PAR:SEL '{name}',fast")
        else:
            self.write(f"CALC{measurements[name].cnum}:PAR:SEL '{name}'")



from __future__ import annotations

import dataclasses
from abc import ABC, abstractmethod

import numpy as np

import skrf


class SweepSection(ABC):
    @abstractmethod
    def get_hz(self) -> list[float]:
        ''' List of hz represented by this section after applying mask '''
        pass

    def get_raw_hz(self) -> list[float]:
        ''' List of hz fetched from the instrument before applying mask '''
        return self.get_hz()

    @abstractmethod
    def apply_8510(self, hp8510c):
        pass

    def mask_8510(self, network : skrf.Network) -> skrf.Network:
        return network


@dataclasses.dataclass
class LinearBuiltinSweepSection(SweepSection):
    hz_min: float
    hz_max: float
    n_points: int

    def get_hz(self) -> list[float]:
        return np.linspace(self.hz_min, self.hz_max, self.n_points)

    def apply_8510(self, hp8510c):
        hp8510c._set_instrument_step_state(self.hz_min, self.hz_max, self.n_points)


@dataclasses.dataclass
class LinearMaskedSweepSection(SweepSection):
    hz_min: float
    hz_max: float
    n_points: int
    mask: list[bool]

    def get_hz(self) -> list[float]:
        return self.get_raw_hz()[self.mask]

    def get_raw_hz(self) -> list[float]:
        return np.linspace(self.hz_min, self.hz_max, self.n_points)

    def apply_8510(self, hp8510c):
        hp8510c._set_instrument_step_state(self.hz_min, self.hz_max, self.n_points)

    def mask_8510(self, network : skrf.Network) -> skrf.Network:
        return network[self.mask]


@dataclasses.dataclass
class LinearCustomSweepSection(SweepSection):
    hz_min: float
    hz_max: float
    n_points: int

    def get_hz(self) -> list[float]:
        return np.linspace(self.hz_min, self.hz_max, self.n_points)

    def get_raw_hz(self) -> list[float]:
        if self.n_points==1:
            return [self.hz_min, self.hz_min+1]
        return np.linspace(self.hz_min, self.hz_max, self.n_points)

    def apply_8510(self, hp8510c):
        assert self.n_points <= 792
        if self.n_points==1:
            hp8510c._set_instrument_step_state(self.hz_min, self.hz_max, 2)
            return
        hp8510c._set_instrument_step_state(self.hz_min, self.hz_max, self.n_points)

    def mask_8510(self, network : skrf.Network) -> skrf.Network:
        if self.n_points==1:
            return network[0]
        return network


@dataclasses.dataclass
class RandomSweepSection(SweepSection):
    hz_list: list[float]

    def get_hz(self) -> list[float]:
        return self.hz_list

    def get_raw_hz(self) -> list[float]:
        ''' List of hz fetched from the instrument before applying mask '''
        if len(self.hz_list)==1:
            return [self.hz_list[0], self.hz_list[0]+1]  # 8510 treats length 1 like length 2
        return self.get_hz()

    def apply_8510(self, hp8510c):
        hp8510c._set_instrument_cwstep_state(self.hz_list)

    def mask_8510(self, network : skrf.Network) -> skrf.Network:
        if len(self.hz_list)==1:
            return network[0]
        return network

def _sweep_sectionsfrom_hz(hz) -> list[SweepSection]:
    """
    Take a list of hz, return a SweepPlan, a list of 8510C compatible
    SweepSections that can be executed to cover the original list of frequencies.
    Limitations: No overlapping linear sweeps. No log sweeps.
    """
    hz = np.array(sorted(hz))
    n = len(hz)
    misfits = []  # hz that don't fit into a linear sweep
    growing_window = []
    growing_window_d = -1
    sweep_sections = []

    def finalize_window(growing_window, misfits, sweep_sections):
        """ When a growing_window has grown as far as it can, finalize_window is called to
        turn it into a sweep section + misfits."""
        # Runt windows aren't really linear sweeps -- we should just add their points to the misfit pile
        if len(growing_window) <= 2:
            misfits.extend(growing_window)
            growing_window.clear()
            return
        # Certain window lengths are preferred. Try to use these as much as possible
        def try_builtin_window_len(builtin_len):
            while len(growing_window) >= builtin_len:
                chunk = growing_window[0:builtin_len]
                del growing_window[0:builtin_len]
                ns = LinearBuiltinSweepSection(
                    hz_min=chunk[0], hz_max=chunk[-1], n_points=len(chunk)
                )
                sweep_sections.append(ns)

        try_builtin_window_len(801)
        try_builtin_window_len(401)
        # try_builtin_window_len(201) # List sweeps beat a succession of small windows
        # try_builtin_window_len(101)
        # try_builtin_window_len(51)

        # Remainder get a custom linear window:
        while len(growing_window):
            chunk = growing_window[0:792]
            del growing_window[0:792]
            ns = LinearCustomSweepSection(
                hz_min=chunk[0], hz_max=chunk[-1], n_points=len(chunk)
            )
            assert len(chunk) == len(ns.get_hz())
            assert np.allclose(chunk, ns.get_hz())
            sweep_sections.append(ns)

    # Start growing windows!
    for i in range(n):
        wl = len(growing_window)
        if wl == 0:
            growing_window.append(hz[i])
        if wl == 1:
            growing_window.append(hz[i])
            growing_window_d = hz[i] - hz[i - 1]
        if wl >= 2:
            i_fits_in_growing_window = (
                abs((hz[i] - hz[i - 1]) - growing_window_d) < 0.5
            )  # Hz tolerance
            if i_fits_in_growing_window:
                growing_window.append(hz[i])
            else:
                finalize_window(growing_window, misfits, sweep_sections)
                growing_window.append(hz[i])
    finalize_window(growing_window, misfits, sweep_sections)
    while len(misfits):
        hz = misfits[:29]
        sweep_sections.append(RandomSweepSection(hz_list=hz))
        del misfits[:29]
    return sweep_sections

class SweepPlan:
    """
    The user requests a big sweep with different spacings in different frequency
    blocks, but the HP8510 instrument only supports smaller sweeps with fixed
    spacing. What to do? Break up the big sweep, naturally.

    SweepPlan.from_hz(array_of_hz)
      ->
        SweepPlan
            SweepSection
            SweepSection
            SweepSection

    Each SweepSection represents a sweep the instrument can actually perform.
    Together, the SweepSections in a SweepPlan satisfy the user's request.

    There is room for cleverness: conducting an 800 point sweep by running an
    801 point sweep and discarding a point saved 20 minutes in a recent script.
    """

    _sections: list[SweepSection]

    def __init__(self, sections):
        self._sections = sections

    @classmethod
    def from_hz(cls, hz : list[float]):
        sweep_sections = _sweep_sectionsfrom_hz(hz)
        plan = SweepPlan(sweep_sections)
        assert plan._matches_f_list(hz)
        return plan

    def get_sections(self):
        return self._sections

    def get_hz(self) -> list[float]:
        """Get a list of all frequency points in the entire sweep plan."""
        ret = []
        for s in self._sections:
            ret.extend(s.get_hz())
        return ret

    def _matches_f_list(self, golden_hz : list[float]):
        """
        Returns True iff the frequencies this SweepPlan intends to sweep
        equal those in the list golden_hz.
        """
        plan_hz = np.array(sorted(self.get_hz()))
        if len(golden_hz) != len(plan_hz):
            print("Planner output has different length.")
            return False
        good = True
        for h in golden_hz:
            if not np.any(np.isclose(h, plan_hz)):
                print(f"In original list but not plan: {h}")
                good = False
        for ph in plan_hz:
            if not np.any(np.isclose(ph, golden_hz)):
                print(f"In plan but not in original list: {h}")
                good = False
        if not np.allclose(golden_hz, plan_hz):
            print("All were not close.")
            good = False
        return good

    @classmethod
    def from_ssn(cls, hz_start, hz_stop, n):
        return cls.from_hz(np.linspace(hz_start, hz_stop, n))



import sys

import pytest

try:
    from skrf.vi.vna import vna
except ImportError:
    pass

if "matplotlib" not in sys.modules:
    pytest.skip(allow_module_level=True)

@pytest.mark.parametrize(
    "cmd,kwargs,expected",
    [
        ("*IDN?", {}, "*IDN?"),
        ("SENS<self:cnum>", {}, "SENS1"),
        ("SENS<self:cnum>:STAR?", {}, "SENS1:STAR?"),
        ("SENS<self:cnum>:STAR <arg>", {"arg": 100}, "SENS1:STAR 100"),
    ],
)
def test_format_cmd(cmd, kwargs, expected):
    class TestDevice:
        cnum = 1
        name = "channel"

    assert vna._format_cmd(cmd, self=TestDevice(), **kwargs) == expected


def test_vna_add_channel_support():
    class TestVNA(vna.VNA):
        pass

    assert not hasattr(TestVNA, "create_channel")
    assert not hasattr(TestVNA, "delete_channel")

    TestVNA._add_channel_support()

    assert hasattr(TestVNA, "create_channel")
    assert hasattr(TestVNA, "delete_channel")


def test_vna_create_delete_channels():
    class TestVNA(vna.VNA):
        class Channel(vna.Channel):
            pass

        def __init__(self):
            pass

    TestVNA._add_channel_support()
    instr = TestVNA()

    assert hasattr(instr, "create_channel")
    assert hasattr(instr, "delete_channel")

    instr.create_channel(1, "channel 1")
    assert hasattr(instr, "ch1")
    instr.delete_channel(1)
    assert not hasattr(instr, "ch1")



"""
mathFunctions (:mod:`skrf.mathFunctions`)
=============================================


Provides commonly used mathematical functions.

Mathematical Constants
----------------------
Some convenient constants are defined in the :mod:`skrf.constants` module.

Complex Component Conversion
---------------------------------
.. autosummary::
        :toctree: generated/

        complex_2_reim
        complex_2_magnitude
        complex_2_db
        complex_2_db10
        complex_2_radian
        complex_2_degree
        complex_2_magnitude
        complex_2_quadrature

Phase Unwrapping
--------------------------------
.. autosummary::
        :toctree: generated/

        unwrap_rad
        sqrt_phase_unwrap



Unit Conversion
--------------------------------
.. autosummary::
        :toctree: generated/

        radian_2_degree
        degree_2_radian
        np_2_db
        db_2_np

Scalar-Complex Conversion
---------------------------------
These conversions are useful for wrapping other functions that don't
support complex numbers.

.. autosummary::
        :toctree: generated/

        complex2Scalar
        scalar2Complex


Special Functions
---------------------------------
.. autosummary::
        :toctree: generated/

        dirac_delta
        neuman
        null
        cross_ratio

Various Utility Functions
--------------------------
.. autosummary::
        :toctree: generated/

        psd2TimeDomain
        rational_interp
        ifft
        irfft
        is_square
        is_symmetric
        is_Hermitian
        is_positive_definite
        is_positive_semidefinite
        get_Hermitian_transpose


"""
from __future__ import annotations

from typing import Callable

import numpy as np
from numpy import imag, pi, real, unwrap
from scipy import signal

from .constants import ALMOST_ZERO, EIG_COND, EIG_MIN, INF, LOG_OF_NEG, NumberLike


# simple conversions
def complex_2_magnitude(z: NumberLike):
    """
    Return the magnitude of the complex argument.

    Parameters
    ----------
    z : number or array_like
        A complex number or sequence of complex numbers

    Returns
    -------
    mag : ndarray or scalar

    """
    return np.abs(z)


def complex_2_db(z: NumberLike):
    r"""
    Return the magnitude in dB of a complex number (as :math:`20\log_{10}(|z|)`)..

    The magnitude in dB is defined as :math:`20\log_{10}(|z|)`
    where :math:`z` is a complex number.

    Parameters
    ----------
    z : number or array_like
        A complex number or sequence of complex numbers

    Returns
    -------
    mag20dB : ndarray or scalar
    """
    return magnitude_2_db(np.abs(z))


def complex_2_db10(z: NumberLike):
    r"""
    Return the magnitude in dB of a complex number (as :math:`10\log_{10}(|z|)`).

    The magnitude in dB is defined as :math:`10\log_{10}(|z|)`
    where :math:`z` is a complex number.

    Parameters
    ----------
    z : number or array_like
        A complex number or sequence of complex numbers

    Returns
    -------
    mag10dB : ndarray or scalar
    """
    return mag_2_db10(np.abs(z))


def complex_2_radian(z: NumberLike):
    """
    Return the angle complex argument in radian.

    Parameters
    ----------
    z : number or array_like
        A complex number or sequence of complex numbers

    Returns
    -------
    ang_rad : ndarray or scalar
        The counterclockwise angle from the positive real axis on the complex
        plane in the range ``(-pi, pi]``, with dtype as numpy.float64.
    """
    return np.angle(z)


def complex_2_degree(z: NumberLike):
    """
    Returns the angle complex argument in degree.

    Parameters
    ----------
    z : number or array_like
        A complex number or sequence of complex numbers

    Returns
    -------
    ang_deg : ndarray or scalar
    """
    return np.angle(z, deg=True)


def complex_2_quadrature(z: NumberLike):
    r"""
    Take a complex number and returns quadrature, which is (length, arc-length from real axis)

    Arc-length is calculated as :math:`|z| \arg(z)`.

    Parameters
    ----------
    z : number or array_like
        A complex number or sequence of complex numbers

    Returns
    -------
    mag : array like or scalar
        magnitude (length)
    arc_length : array like or scalar
        arc-length from real axis: angle*magnitude
    """
    return (np.abs(z), np.angle(z)*np.abs(z))


def complex_2_reim(z: NumberLike):
    """
    Return real and imaginary parts of a complex number.

    Parameters
    ----------
    z : number or array_like
        A complex number or sequence of complex numbers

    Returns
    -------
    real : array like or scalar
        real part of input
    imag : array like or scalar
        imaginary part of input
    """
    return (np.real(z), np.imag(z))


def complex_components(z: NumberLike):
    """
    Break up a complex array into all possible scalar components.

    Parameters
    ----------
    z : number or array_like
        A complex number or sequence of complex numbers

    Returns
    -------
    c_real : array like or scalar
        real part
    c_imag : array like or scalar
        imaginary part
    c_angle : array like or scalar
        angle in degrees
    c_mag : array like or scalar
        magnitude
    c_arc : array like or scalar
        arclength from real axis, angle*magnitude
    """
    return (*complex_2_reim(z), np.angle(z,deg=True), *complex_2_quadrature(z))


def magnitude_2_db(z: NumberLike, zero_nan: bool = True):
    """
    Convert linear magnitude to dB.

    Parameters
    ----------
    z : number or array_like
        A complex number or sequence of complex numbers
    zero_nan : bool, optional
        Replace NaN with zero. The default is True.

    Returns
    -------
    z : number or array_like
       Magnitude in dB given by 20*log10(|z|)
    """
    out = 20 * np.log10(z)
    if zero_nan:
        return np.nan_to_num(out, nan=LOG_OF_NEG, neginf=-np.inf)
    return out

mag_2_db = magnitude_2_db


def mag_2_db10(z: NumberLike, zero_nan:bool = True):
    """
    Convert linear magnitude to dB.

    Parameters
    ----------
    z : array_like
        A complex number or sequence of complex numbers
    zero_nan : bool, optional
        Replace NaN with zero. The default is True.

    Returns
    -------
    z : array_like
       Magnitude in dB given by 10*log10(|z|)
    """
    out = 10 * np.log10(z)
    if zero_nan:
        return np.nan_to_num(out, nan=LOG_OF_NEG, neginf=-np.inf)
    return out


def db_2_magnitude(z: NumberLike):
    """
    Convert dB to linear magnitude.

    Parameters
    ----------
    z : number or array_like
        A complex number or sequence of complex numbers

    Returns
    -------
    z : number or array_like
        10**((z)/20) where z is a complex number
    """
    return 10**((z)/20.)

db_2_mag = db_2_magnitude


def db10_2_mag(z: NumberLike):
    """
    Convert dB to linear magnitude.

    Parameters
    ----------
    z : array_like
        A complex number or sequence of complex numbers

    Returns
    -------
    z : array_like
        10**((z)/10) where z is a complex number
    """
    return 10**((z)/10.)


def magdeg_2_reim(mag: NumberLike, deg: NumberLike):
    """
    Convert linear magnitude and phase (in deg) arrays into a complex array.

    Parameters
    ----------
    mag : number or array_like
        A complex number or sequence of real numbers
    deg : number or array_like
        A complex number or sequence of real numbers

    Returns
    -------
    z : array_like
        A complex number or sequence of complex numbers

    """
    return mag*np.exp(1j*deg*pi/180.)

def dbdeg_2_reim(db: NumberLike, deg: NumberLike):
    """
    Converts dB magnitude and phase (in deg) arrays into a complex array.

    Parameters
    ----------
    db : number or array_like
        A realnumber or sequence of real numbers
    deg : number or array_like
        A real number or sequence of real numbers

    Returns
    -------
    z : array_like
        A complex number or sequence of complex numbers
    """
    return magdeg_2_reim(db_2_magnitude(db), deg)


def db_2_np(db: NumberLike):
    """
    Converts a value in decibel (dB) to neper (Np).

    Parameters
    ----------
    db : number or array_like
        A real number or sequence of real numbers

    Returns
    -------
    np : number or array_like
        A real number of sequence of real numbers
    """
    return (np.log(10)/20) * db


def np_2_db(x: NumberLike):
    """
    Converts a value in Nepers (Np) to decibel (dB).

    Parameters
    ----------
    np : number or array_like
        A real number or sequence of real numbers

    Returns
    -------
    db : number or array_like
        A real number of sequence of real numbers
    """
    return 20/np.log(10) * x


def radian_2_degree(rad: NumberLike):
    """
    Convert angles from radians to degrees.

    Parameters
    ----------
    rad : number or array_like
        Angle in radian

    Returns
    -------
    deg : number or array_like
        Angle in degree
    """
    return (rad)*180/pi


def degree_2_radian(deg: NumberLike):
    """
    Convert angles from degrees to radians.

    Parameters
    ----------
    deg : number or array_like
        Angle in radian

    Returns
    -------
    rad : number or array_like
        Angle in degree
    """
    return (deg)*pi/180.


def feet_2_meter(feet: NumberLike = 1):
    """
    Convert length in feet to meter.

    1 foot is equal to 0.3048 meters.

    Parameters
    ----------
    feet : number or array-like, optional
        length in feet. Default is 1.

    Returns
    -------
    meter: number or array-like
        length in meter

    See Also
    --------
    meter_2_feet
    """
    return 0.3048*feet

def meter_2_feet(meter: NumberLike = 1):
    """
    Convert length in meter to feet.

    1 meter is equal to 0.3.28084 feet.

    Parameters
    ----------
    meter : number or array-like, optional
        length in meter. Default is 1.

    Returns
    -------
    feet : number or array-like
        length in feet

    See Also
    --------
    feet_2_meter
    """
    return 3.28084*meter


def db_per_100feet_2_db_per_100meter(db_per_100feet: NumberLike = 1):
    """
    Convert attenuation values given in dB/100ft to dB/100m.

    db_per_100meter = db_per_100feet * rf.meter_2_feet()

    Parameters
    ----------
    db_per_100feet : number or array-like, optional
        Attenuation in dB/ 100 ft. Default is 1.

    Returns
    -------
    db_per_100meter : number or array-like
        Attenuation in dB/ 100 m

    See Also
    --------
    meter_2_feet
    feet_2_meter
    np_2_db
    db_2_np
    """
    return db_per_100feet * 100 / feet_2_meter(100)


def unwrap_rad(phi: NumberLike):
    """
    Unwraps a phase given in radians.

    Parameters
    ----------
    phi : number or array_like
        phase in radians

    Returns
    -------
    phi : number of array_like
        unwrapped phase in radians
    """
    return unwrap(phi, axis=0)


def sqrt_known_sign(z_squared: NumberLike, z_approx: NumberLike):
    """
    Return the square root of a complex number, with sign chosen to match `z_approx`.

    Parameters
    ----------
    z_squared : number or array-like
        the complex to be square-rooted
    z_approx : number or array-like
        the approximate value of z. sign of z is chosen to match that of
        z_approx

    Returns
    -------
    z : number, array-like (same type as z_squared)
        square root of z_squared.
    """
    z = np.sqrt(z_squared)
    return np.where(
        np.sign(np.angle(z)) == np.sign(np.angle(z_approx)),
        z, z.conj())


def find_correct_sign(z1: NumberLike, z2: NumberLike, z_approx: NumberLike):
    r"""
    Create new vector from z1, z2 choosing elements with sign matching z_approx.

    This is used when you have to make a root choice on a complex number.
    and you know the approximate value of the root.

    .. math::

        z1,z2 = \pm \sqrt(z^2)


    Parameters
    ----------
    z1 : array-like
        root 1
    z2 : array-like
        root 2
    z_approx : array-like
        approximate answer of z

    Returns
    -------
    z3 : np.array
        array built from z1 and z2 by
        z1 where sign(z1) == sign(z_approx), z2 else

    """
    return np.where(
    np.sign(np.angle(z1)) == np.sign(np.angle(z_approx)),z1, z2)


def find_closest(z1: NumberLike, z2: NumberLike, z_approx: NumberLike):
    """
    Return z1 or z2  depending on which is  closer to z_approx.

    Parameters
    ----------
    z1 : array-like
        root 1
    z2 : array-like
        root 2
    z_approx : array-like
        approximate answer of z

    Returns
    -------
    z3 : np.array
        array built from z1 and z2

    """
    z1_dist = abs(z1-z_approx)
    z2_dist = abs(z2-z_approx)

    return np.where(z1_dist<z2_dist,z1, z2)

def sqrt_phase_unwrap(z: NumberLike):
    r"""
    Take the square root of a complex number with unwrapped phase.

    This idea came from Lihan Chen.

    .. math::

        \sqrt{|z|} \exp( \arg_{unwrap}(z) / 2 )


    Parameters
    ----------
    z : number or array_like
        A complex number or sequence of complex numbers

    Returns
    -------
    z : number of array_like
        A complex number or sequence of complex numbers
    """
    return np.sqrt(abs(z))*\
            np.exp(0.5*1j*unwrap_rad(complex_2_radian(z)))


# mathematical functions
def dirac_delta(x: NumberLike):
    r"""
    Calculate Dirac function.

    Dirac function :math:`\delta(x)` defined as :math:`\delta(x)=1` if x=0,
    0 otherwise.

    Parameters
    ----------
    x : number of array_like
        A real number or sequence of real numbers

    Returns
    -------
    delta : number of array_like
        1 or 0

    References
    ----------
    https://en.wikipedia.org/wiki/Dirac_delta_function

    """
    return (x==0)*1. + (x!=0)*0.


def neuman(x: NumberLike):
    r"""
    Calculate Neumans number.

    It is defined as:

    .. math::

        2 - \delta(x)

    where :math:`\delta` is the Dirac function.

    Parameters
    ----------
    x : number or array_like
        A real number or sequence of real numbers

    Returns
    -------
    y : number or array_like
        A real number or sequence of real numbers

    See Also
    --------
    dirac_delta
    """
    return 2. - dirac_delta(x)


def null(A: np.ndarray, eps: float = 1e-15):
    """
    Calculate the null space of matrix A.

    Parameters
    ----------
    A : array_like
    eps : float

    Returns
    -------
    null_space : array_like

    References
    ----------
    https://scipy-cookbook.readthedocs.io/items/RankNullspace.html
    https://stackoverflow.com/questions/5889142/python-numpy-scipy-finding-the-null-space-of-a-matrix
    """
    u, s, vh = np.linalg.svd(A)
    null_space = np.compress(s <= eps, vh, axis=0)
    return null_space.T


def inf_to_num(x: NumberLike):
    """
    Convert inf and -inf's to large numbers.

    Parameters
    ------------
    x : array-like or number
        the input array or number

    Returns
    -------
    x : Number of array_like
        Input without with +/- inf replaced by large numbers
    """
    x = np.nan_to_num(x, nan=np.nan, posinf=INF, neginf=-1*INF)
    return x


def cross_ratio(a: NumberLike, b: NumberLike, c: NumberLike, d:NumberLike):
    r"""
    Calculate the cross ratio of a quadruple of distinct points on the real line.


    The cross ratio is defined as:


    .. math::

        r = \frac{ (a-b)(c-d) }{ (a-d)(c-b) }



    Parameters
    ----------
    a,b,c,d :  array-like or number

    Returns
    -------
    r : array-like or number

    References
    ----------
    https://en.wikipedia.org/wiki/Cross-ratio

    """
    return ((a-b)*(c-d))/((a-d)*(c-b))


def complexify(f: Callable, name: str = None):
    """
    Make a function f(scalar) into f(complex).

    If `f(x)` then it returns `f_c(z) = f(real(z)) + 1j*f(imag(z))`

    If the real/imag arguments are not first, then you may specify the
    name given to them as kwargs.

    Parameters
    ----------
    f : Callable
        function of real variable
    name : string, optional
        name of the real/imag argument names if they are not first

    Returns
    -------
    f_c : Callable
        function of a complex variable

    Examples
    --------
    >>> def f(x): return x
    >>> f_c = rf.complexify(f)
    >>> z = 0.2 -1j*0.3
    >>> f_c(z)

    """
    def f_c(z, *args, **kw):
        if name is not None:
            kw_re = {name: real(z)}
            kw_im = {name: imag(z)}
            kw_re.update(kw)
            kw_im.update(kw)
            return f(*args, **kw_re) + 1j*f(*args, **kw_im)
        else:
            return f(real(z), *args,**kw) + 1j*f(imag(z), *args, **kw)
    return f_c



# old functions just for reference
def complex2Scalar(z: NumberLike):
    """
    Serialize a list/array of complex numbers

    Parameters
    ----------
    z : number or array_like
        A complex number or sequence of complex numbers

    Returns
    -------
    re_im : array_like
        produce the following array for an input z:
        z[0].real, z[0].imag, z[1].real, z[1].imag, etc.

    See Also
    --------
    scalar2Complex
    """
    z = np.array(z)
    re_im = []
    for k in z:
        re_im.append(np.real(k))
        re_im.append(np.imag(k))
    return np.array(re_im).flatten()

def scalar2Complex(s: NumberLike):
    """
    Unserialize a list/array of real and imag numbers into a complex array.

    Inverse of :func:`complex2Scalar`.

    Parameters
    ----------
    s : array_like
        an array with real and imaginary parts ordered as:
        z[0].real, z[0].imag, z[1].real, z[1].imag, etc.

    Returns
    -------
    z : Number or array_like
        A complex number or sequence of complex number

    See Also
    --------
    complex2Scalar
    """
    s = np.array(s)
    z = []

    for k in range(0,len(s),2):
        z.append(s[k] + 1j*s[k+1])
    return np.array(z).flatten()


def flatten_c_mat(s: NumberLike, order: str = 'F'):
    """
    Take a 2D (mxn) complex matrix and serialize and flatten it.

    by default (using order='F') this generates the following
    from a  2x2

    [s11,s12;s21,s22]->[s11re,s11im,s21re,s12im, ...]

    Parameters
    ------------
    s : ndarray
        input 2D array
    order : sting, optional
        either 'F' or 'C', for the order of flattening
    """
    return complex2Scalar(s.flatten(order='F'))


def rand_c(*size) -> np.ndarray:
    """
    Creates a complex random array of shape s.

    The bounds on real and imaginary values are (-1,1)

    Parameters
    -----------
    s : list-like
        shape of array

    Examples
    ---------
    >>> x = rf.rand_c(2,2)
    """
    rng = np.random.default_rng()
    return 1-2*rng.random(size) + \
        1j-2j*rng.random(size)



def psd2TimeDomain(f: np.ndarray, y: np.ndarray, windowType: str = 'hamming'):
    """
    Convert a one sided complex spectrum into a real time-signal.

    Parameters
    ----------
    f : list or np.ndarray
        frequency array
    y : list of np.ndarray
        complex PSD array
    windowType: string
        windowing function, defaults to 'hamming''

    Returns
    -------
    timeVector : array_like
        inverse units of the input variable f,
    signalVector : array_like

    Note
    ----
    If spectrum is not baseband then, `timeSignal` is modulated by `exp(t*2*pi*f[0])`.
    So keep in mind units. Also, due to this, `f` must be increasing left to right.
    """


    # apply window function
    # make sure windowType exists in scipy.signal
    if callable(getattr(signal, windowType)) and (windowType != 'rect' ):
        window = getattr(signal, windowType)(len(f))
        y = y * window

    #create other half of spectrum
    spectrum = (np.hstack([np.real(y[:0:-1]),np.real(y)])) + \
            1j*(np.hstack([-np.imag(y[:0:-1]),np.imag(y)]))

    # do the transform
    df = abs(f[1]-f[0])
    T = 1./df
    timeVector = np.linspace(-T/2.,T/2,2*len(f)-1)
    signalVector = np.fft.ifftshift(np.fft.ifft(np.fft.ifftshift(spectrum)))

    #the imaginary part of this signal should be from fft errors only,
    signalVector= np.real(signalVector)
    # the response of frequency shifting is
    # exp(1j*2*pi*timeVector*f[0])
    # but I would have to manually undo this for the inverse, which is just
    # another variable to require. The reason you need this is because
    # you can't transform to a bandpass signal, only a lowpass.
    #
    return timeVector, signalVector


def rational_interp(
        x: np.ndarray,
        y: np.ndarray,
        d: int = 4,
        epsilon: float = 1e-9,
        axis: int = 0,
        assume_sorted: bool = False) -> Callable:
    """
    Interpolates function using rational polynomials of degree `d`.

    Interpolating function is singular when xi is exactly one of the
    original x points. If xi is closer than epsilon to one of the original points,
    then the value at that points is returned instead.

    Implementation is based on [#]_.

    Parameters
    ----------
    x : np.ndarray
    y : np.ndarray
    d : int, optional
        order of the polynomial, by default 4
    epsilon : float, optional
        numerical tolerance, by default 1e-9
    axis : int, optional
        axis to operate on, by default 0
    assume_sorted : bool, optional
        If False, values of x can be in any order and they are sorted first.
        If True, x has to be an array of monotonically increasing values.

    Returns
    -------
    fx : Callable
        Interpolate function

    Raises
    ------
    NotImplementedError
        if axis != 0.

    References
    ------------
    .. [#] M. S. Floater and K. Hormann, "Barycentric rational interpolation with no poles and high rates of
    approximation," Numer. Math., vol. 107, no. 2, pp. 315-331, Aug. 2007
    """
    if axis != 0:
        raise NotImplementedError("Axis other than 0 is not implemented")

    if not assume_sorted:
        sort_indices = np.argsort(x, axis=axis)
        x = x[sort_indices]
        y = y[sort_indices]

    n = len(x)
    if n <= d:
        raise ValueError('Not enough x-axis points')

    w = np.zeros(n)
    # Scaling to give close to 1 weights
    hd = (x[n//2] - x[n//2-1])**d
    for k in range(n):
        for i in range(max(0,k-d), min(k+1, n-d)):
            p, xk = hd, x[k]
            for j in range(i,min(n,i+d+1)):
                if j == k:
                    continue
                p /= (xk - x[j])
            if i % 2 == 1:
                w[k] -= p
            else:
                w[k] += p

    # Add dimensions to match y shape
    w_shape = [1]*len(y.shape)
    w_shape[0] = -1
    w = w.reshape(w_shape)

    def fx(xi):
        # The method will divide by zero if new x value is exactly existing x value.
        # To avoid this we need to check for too close values and replace them with
        # y value at that position.
        idx = np.searchsorted(x, xi)
        idx[idx == len(x)] = len(x) - 1
        nearest_idx = np.where(np.abs(x[idx] - xi) < epsilon)[0]
        nearest_value = y[idx[nearest_idx]]

        xi = xi.reshape(*w_shape)
        with np.errstate(divide='ignore', invalid='ignore'):
            assert axis == 0
            tmp = [w[i] / (xi - x[i]) for i in range(n)]
            v = sum(y[i] * tmp[i] for i in range(n)) / sum(tmp)

        # Fix divide by zero errors
        for e, i in enumerate(nearest_idx):
            v[i] = nearest_value[e]

        return v

    return fx

def ifft(x: np.ndarray) -> np.ndarray:
    """
    Transforms S-parameters to time-domain bandpass.

    Parameters
    ----------
    x : array_like
        S-parameters vs frequency array.

    Returns
    -------
    X : array_like
        Fourier transformed array

    See Also
    --------
    irfft
    """
    return np.fft.fftshift(np.fft.ifft(np.fft.ifftshift(x, axes=0), axis=0), axes=0)


def irfft(x: np.ndarray, n:int = None) -> np.ndarray:
    """
    Transforms S-parameters to time-domain, assuming complex conjugates for
    values corresponding to negative frequencies.

    Parameters
    ----------
    x : array_like
        S-parameters vs frequency array.
    n : int, optional
        n parameter passed to :func:`numpy.fft.irfft`. Defaults to None.

    Returns
    -------
    X : array_like
        Fourier transformed array

    See Also
    --------
    ifft
    """
    return np.fft.fftshift(np.fft.irfft(x, axis=0, n=n), axes=0)


def is_square(mat: np.ndarray) -> bool:
    """
    Tests whether mat is a square matrix.

    Parameters
    ----------
    mat : np.ndarray
        Matrix to test for being square

    Returns
    -------
    res : boolean

    See Also
    --------
    is_unitary
    is_symmetric
    """
    return mat.shape[0] == mat.shape[1]


def is_unitary(mat: np.ndarray, tol: float = ALMOST_ZERO) -> bool:
    """
    Tests mat for unitariness.

    Parameters
    ----------
    mat : np.ndarray
        Matrix to test for unitariness
    tol : float
        Absolute tolerance. Defaults to :data:`ALMOST_ZERO`

    Returns
    -------
    res : boolean or array of boolean

    See Also
    --------
    is_square
    is_symmetric

    """
    if not is_square(mat):
        return False
    return np.allclose(get_Hermitian_transpose(mat) @ mat,
                        np.identity(mat.shape[0]), atol=tol)


def is_symmetric(mat: np.ndarray, tol: int = ALMOST_ZERO) -> bool:
    """
    Tests mat for symmetry.

    Parameters
    ----------
    mat : np.ndarray
        Matrix to test for symmetry
    tol : float, optional
        Absolute tolerance. Defaults to :data:`ALMOST_ZERO`

    Returns
    -------
    res : boolean or array of boolean

    See Also
    --------
    is_square
    is_unitary
    """
    if not is_square(mat):
        return False
    return np.allclose(mat, mat.transpose(), atol=tol)


def get_Hermitian_transpose(mat: np.ndarray) -> np.ndarray:
    """
    Returns the conjugate transpose of mat.

    Parameters
    ----------
    mat : np.ndarray
        Matrix to compute the conjugate transpose of

    Returns
    -------
    mat : np.ndarray

    """
    return mat.transpose().conjugate()


def is_Hermitian(mat: np.ndarray, tol: float = ALMOST_ZERO) -> bool:
    """
    Tests whether mat is Hermitian.

    Parameters
    ----------
    mat : np.ndarray
        Matrix to test for being Hermitian
    tol : float
        Absolute tolerance

    Returns
    -------
    res : boolean

    See Also
    --------
    is_positive_definite
    is_positive_semidefinite
    """
    if not is_square(mat):
        return False
    return np.allclose(mat, get_Hermitian_transpose(mat), atol=tol)


def is_positive_definite(mat: np.ndarray, tol: float = ALMOST_ZERO) -> bool:
    """
    Tests mat for positive definiteness.

    Verifying that
    (1) mat is symmetric
    (2) it's possible to compute the Cholesky decomposition of mat.

    Parameters
    ----------
    mat : np.ndarray
        Matrix to test for positive definiteness
    tol : float, optional
        Absolute tolerance. Defaults to :data:`ALMOST_ZERO`

    Returns
    -------
    res : bool or array of bool

    See Also
    --------
    is_Hermitian
    is_positive_semidefinite
    """
    if not is_Hermitian(mat, tol=tol):
        return False
    try:
        np.linalg.cholesky(mat)
        return True
    except np.linalg.LinAlgError:
        return False


def is_positive_semidefinite(mat: np.ndarray, tol: float = ALMOST_ZERO) -> bool:
    """
    Tests mat for positive semidefiniteness.

    Checking whether all eigenvalues of mat are nonnegative within a certain tolerance

    Parameters
    ----------
    mat : np.ndarray
        Matrix to test for positive semidefiniteness
    tol : float, optional
        Absolute tolerance in determining nonnegativity due to loss of precision
        when computing the eigenvalues of mat. Defaults to :data:`ALMOST_ZERO`

    Returns
    -------
    res : bool or array of bool

    See Also
    --------
    is_Hermitian
    is_positive_definite
    """
    if not is_Hermitian(mat):
        return False
    try:
        v = np.linalg.eigvalsh(mat)
    except np.linalg.LinAlgError:
        return False
    return np.all(v > -tol)

def rsolve(A: np.ndarray, B: np.ndarray) -> np.ndarray:
    r"""Solves x @ A = B.

    Calls numpy.linalg.solve with transposed matrices.

    Same as B @ np.linalg.inv(A) but avoids calculating the inverse and
    should be numerically slightly more accurate.

    Input should have dimension of similar to (nfreqs, nports, nports).

    Parameters
    ----------
    A : np.ndarray
    B : np.ndarray

    Returns
    -------
    x : np.ndarray
    """
    return np.transpose(np.linalg.solve(np.transpose(A, (0, 2, 1)).conj(),
            np.transpose(B, (0, 2, 1)).conj()), (0, 2, 1)).conj()

def nudge_eig(mat: np.ndarray,
              cond: float | None = None,
              min_eig: float | None  = None) -> np.ndarray:
    r"""Nudge eigenvalues with absolute value smaller than
    max(cond * max(eigenvalue), min_eig) to that value.
    Can be used to avoid singularities in solving matrix equations.

    Input should have dimension of similar to (nfreqs, nports, nports).

    Parameters
    ----------
    mat : np.ndarray
        Matrices to nudge
    cond : float, optional
        Minimum eigenvalue ratio compared to the maximum eigenvalue.
        Default value is set by `skrf.constants.EIG_COND`.
    min_eig : float, optional
        Minimum eigenvalue.
        Default value is set by `skrf.constants.EIG_MIN`.
    Returns
    -------
    res : np.ndarray
        Nudged matrices
    """
    # use current constants
    if not cond:
        cond = EIG_COND
    if not min_eig:
        min_eig = EIG_MIN

    # Eigenvalues and vectors
    eigw, eigv = np.linalg.eig(mat)
    # Max eigenvalue for each frequency
    max_eig = np.amax(np.abs(eigw), axis=1)
    # Calculate mask for positions where problematic eigenvalues are
    mask = np.logical_or(np.abs(eigw) < cond * max_eig[:, None], np.abs(eigw) < min_eig)
    if not mask.any():
        # Nothing to do. Return the original array.
        return mat

    mask_cond = cond * np.repeat(max_eig[:, None], mat.shape[-1], axis=-1)[mask]
    mask_min = min_eig * np.ones(mask_cond.shape)
    # Correct the eigenvalues
    eigw[mask] = np.maximum(mask_cond, mask_min)

    # Now assemble the eigendecomposited matrices back
    e = np.zeros_like(mat)
    np.einsum('ijj->ij', e)[...] = eigw
    return rsolve(eigv, eigv @ e)
